"""Methods for overall dataset quality calculations - rows count, a specific values count, etc."""

import dataclasses
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from scipy.stats import chi2_contingency

from evidently.utils.data_operations import DatasetColumns
from evidently.utils.types import ColumnDistribution
from evidently.utils.visualizations import Distribution
from evidently.utils.visualizations import make_hist_for_cat_plot
from evidently.utils.visualizations import make_hist_for_num_plot

MAX_CATEGORIES = 5


def get_rows_count(data: Union[pd.DataFrame, pd.Series]) -> int:
    """Count quantity of rows in  a dataset"""
    return data.shape[0]


@dataclasses.dataclass
class FeatureQualityStats:
    """Class for all features data quality metrics store.

    A type of the feature is stored in `feature_type` field.
    Concrete stat kit depends on the feature type. Is a metric is not applicable - leave `None` value for it.

    Metrics for all feature types:
        - feature type - cat for category, num for numeric, datetime for datetime features
        - count - quantity of a meaningful values (do not take into account NaN values)
        - missing_count - quantity of meaningless (NaN) values
        - missing_percentage - the percentage of the missed values
        - unique_count - quantity of unique values
        - unique_percentage - the percentage of the unique values
        - max - maximum value (not applicable for category features)
        - min - minimum value (not applicable for category features)
        - most_common_value - the most common value in the feature values
        - most_common_value_percentage - the percentage of the most common value
        - most_common_not_null_value - if `most_common_value` equals NaN - the next most common value. Otherwise - None
        - most_common_not_null_value_percentage - the percentage of `most_common_not_null_value` if it is defined.
            If `most_common_not_null_value` is not defined, equals None too.

    Metrics for numeric features only:
        - infinite_count - quantity infinite values (for numeric features only)
        - infinite_percentage - the percentage of infinite values (for numeric features only)
        - percentile_25 - 25% percentile for meaningful values
        - percentile_50 - 50% percentile for meaningful values
        - percentile_75 - 75% percentile for meaningful values
        - mean - the sum of the meaningful values divided by the number of the meaningful values
        - std - standard deviation of the values

    Metrics for category features only:
        - new_in_current_values_count - quantity of new values in the current dataset after the reference
            Defined for reference dataset only.
        - new_in_current_values_count - quantity of values in the reference dataset that not presented in the current
            Defined for reference dataset only.
    """

    # feature type - cat for category, num for numeric, datetime for datetime features
    feature_type: str
    # quantity on
    number_of_rows: int = 0
    count: int = 0
    infinite_count: Optional[int] = None
    infinite_percentage: Optional[float] = None
    missing_count: Optional[int] = None
    missing_percentage: Optional[float] = None
    unique_count: Optional[int] = None
    unique_percentage: Optional[float] = None
    percentile_25: Optional[float] = None
    percentile_50: Optional[float] = None
    percentile_75: Optional[float] = None
    max: Optional[Union[int, float, bool, str]] = None
    min: Optional[Union[int, float, bool, str]] = None
    mean: Optional[float] = None
    most_common_value: Optional[Union[int, float, bool, str]] = None
    most_common_value_percentage: Optional[float] = None
    std: Optional[float] = None
    most_common_not_null_value: Optional[Union[int, float, bool, str]] = None
    most_common_not_null_value_percentage: Optional[float] = None
    new_in_current_values_count: Optional[int] = None
    unused_in_current_values_count: Optional[int] = None

    def is_datetime(self):
        """Checks that the object store stats for a datetime feature"""
        return self.feature_type == "datetime"

    def is_numeric(self):
        """Checks that the object store stats for a numeric feature"""
        return self.feature_type == "num"

    def is_category(self):
        """Checks that the object store stats for a category feature"""
        return self.feature_type == "cat"

    def as_dict(self):
        return {field.name: getattr(self, field.name) for field in dataclasses.fields(FeatureQualityStats)}

    def __eq__(self, other):
        for field in dataclasses.fields(FeatureQualityStats):
            other_field_value = getattr(other, field.name)
            self_field_value = getattr(self, field.name)

            if pd.isnull(other_field_value) and pd.isnull(self_field_value):
                continue

            if not other_field_value == self_field_value:
                return False

        return True


@dataclasses.dataclass
class DataQualityStats:
    rows_count: int
    num_features_stats: Optional[Dict[str, FeatureQualityStats]] = None
    cat_features_stats: Optional[Dict[str, FeatureQualityStats]] = None
    datetime_features_stats: Optional[Dict[str, FeatureQualityStats]] = None
    target_stats: Optional[Dict[str, FeatureQualityStats]] = None
    prediction_stats: Optional[Dict[str, FeatureQualityStats]] = None

    def get_all_features(self) -> Dict[str, FeatureQualityStats]:
        result = {}

        for features in (
            self.target_stats,
            self.prediction_stats,
            self.datetime_features_stats,
            self.cat_features_stats,
            self.num_features_stats,
        ):
            if features is not None:
                result.update(features)

        return result

    def __getitem__(self, item) -> FeatureQualityStats:
        for features in (
            self.target_stats,
            self.prediction_stats,
            self.datetime_features_stats,
            self.cat_features_stats,
            self.num_features_stats,
        ):
            if features is not None and item in features:
                return features[item]

        raise KeyError(item)


def get_features_stats(feature: pd.Series, feature_type: str) -> FeatureQualityStats:
    def get_percentage_from_all_values(value: Union[int, float]) -> float:
        return np.round(100 * value / all_values_count, 2)

    result = FeatureQualityStats(feature_type=feature_type)
    all_values_count = feature.shape[0]

    if not all_values_count > 0:
        # we have no data, return default stats for en empty dataset
        return result
    result.number_of_rows = all_values_count
    result.missing_count = int(feature.isnull().sum())
    result.count = int(feature.count())
    all_values_count = feature.shape[0]
    value_counts = feature.value_counts(dropna=False)
    result.missing_percentage = np.round(100 * result.missing_count / all_values_count, 2)
    unique_count: int = feature.nunique()
    result.unique_count = unique_count
    result.unique_percentage = get_percentage_from_all_values(unique_count)
    result.most_common_value = value_counts.index[0]
    result.most_common_value_percentage = get_percentage_from_all_values(value_counts.iloc[0])

    if result.count > 0 and pd.isnull(result.most_common_value):
        result.most_common_not_null_value = value_counts.index[1]
        result.most_common_not_null_value_percentage = get_percentage_from_all_values(value_counts.iloc[1])

    if feature_type == "num":
        # round most common feature value for numeric features to 1e-5
        if not np.issubdtype(feature, np.number):
            feature = feature.astype(float)
        result.most_common_value = np.round(result.most_common_value, 5)
        result.infinite_count = int(np.sum(np.isinf(feature)))
        result.infinite_percentage = get_percentage_from_all_values(result.infinite_count)
        result.max = np.round(feature.max(), 2)
        result.min = np.round(feature.min(), 2)
        common_stats = dict(feature.describe())
        std = common_stats["std"]
        result.std = np.round(std, 2)
        result.mean = np.round(common_stats["mean"], 2)
        result.percentile_25 = np.round(common_stats["25%"], 2)
        result.percentile_50 = np.round(common_stats["50%"], 2)
        result.percentile_75 = np.round(common_stats["75%"], 2)

    if feature_type == "datetime":
        # cast datetime value to str for datetime features
        result.most_common_value = str(result.most_common_value)
        # cast datetime value to str for datetime features
        result.max = str(feature.max())
        result.min = str(feature.min())

    return result


def calculate_data_quality_stats(
    dataset: pd.DataFrame, columns: DatasetColumns, task: Optional[str]
) -> DataQualityStats:
    result = DataQualityStats(rows_count=get_rows_count(dataset))

    result.num_features_stats = {
        feature_name: get_features_stats(dataset[feature_name], feature_type="num")
        for feature_name in columns.num_feature_names
    }

    result.cat_features_stats = {
        feature_name: get_features_stats(dataset[feature_name], feature_type="cat")
        for feature_name in columns.cat_feature_names
    }

    if columns.utility_columns.date:
        date_list = columns.datetime_feature_names + [columns.utility_columns.date]

    else:
        date_list = columns.datetime_feature_names

    result.datetime_features_stats = {
        feature_name: get_features_stats(dataset[feature_name], feature_type="datetime") for feature_name in date_list
    }

    target_name = columns.utility_columns.target

    if target_name is not None and target_name in dataset:
        result.target_stats = {}

        if task == "classification":
            result.target_stats[target_name] = get_features_stats(dataset[target_name], feature_type="cat")

        else:
            result.target_stats[target_name] = get_features_stats(dataset[target_name], feature_type="num")

    prediction_name = columns.utility_columns.prediction

    if isinstance(prediction_name, str) and prediction_name in dataset:
        result.prediction_stats = {}

        if task == "classification":
            result.prediction_stats[prediction_name] = get_features_stats(dataset[prediction_name], feature_type="cat")

        else:
            result.prediction_stats[prediction_name] = get_features_stats(dataset[prediction_name], feature_type="num")

    return result


@dataclasses.dataclass
class DataQualityPlot:
    bins_for_hist: Dict[str, pd.DataFrame]


class DataQualityGetPlotData:
    def __init__(self) -> None:
        self.period_prefix: Optional[str] = None
        self.curr: Optional[pd.Series] = None
        self.ref: Optional[pd.Series] = None

    def calculate_main_plot(
        self,
        curr: pd.DataFrame,
        ref: Optional[pd.DataFrame],
        feature_name: str,
        feature_type: str,
        merge_small_cat: Optional[int] = MAX_CATEGORIES,
    ):
        if feature_type == "cat" and merge_small_cat is not None:
            if ref is not None:
                ref = ref.copy()
            if merge_small_cat is not None:
                curr, ref = self._transform_cat_data(curr.copy(), ref, feature_name, merge_small_cat)
        curr_data = curr[feature_name].dropna()
        ref_data = None
        if ref is not None:
            ref_data = ref[feature_name].dropna()
        bins_for_hist = None
        if feature_type == "num":
            bins_for_hist = make_hist_for_num_plot(curr_data, ref_data)
            log_ref_data = None
            if ref_data is not None:
                log_ref_data = np.log10(ref_data[ref_data > 0])
            bins_for_hist_log = make_hist_for_num_plot(
                np.log10(curr_data[curr_data > 0]),
                log_ref_data,
            )
            bins_for_hist["current_log"] = bins_for_hist_log["current"]
            if "reference" in bins_for_hist_log.keys():
                bins_for_hist["reference_log"] = bins_for_hist_log["reference"]
        if feature_type == "cat":
            bins_for_hist = make_hist_for_cat_plot(curr_data, ref_data, dropna=True)
        if feature_type == "datetime":
            bins_for_hist = {}
            freq = self._choose_agg_period(feature_name, ref, curr)
            curr_data = curr[feature_name].dt.to_period(freq=freq)
            curr_data = curr_data.value_counts().reset_index()
            curr_data.columns = [feature_name, "number_of_items"]
            curr_data[feature_name] = curr_data[feature_name].dt.to_timestamp()
            bins_for_hist["current"] = curr_data
            if ref is not None:
                ref_data = ref[feature_name].dt.to_period(freq=freq)
                ref_data = ref_data.value_counts().reset_index()
                ref_data.columns = [feature_name, "number_of_items"]
                ref_data[feature_name] = ref_data[feature_name].dt.to_timestamp()
                max_ref_date = ref_data[feature_name].max()
                min_curr_date = curr_data[feature_name].min()
                if max_ref_date == min_curr_date:
                    curr_data, ref_data = self._split_periods(curr_data, ref_data, feature_name)
                bins_for_hist["reference"] = ref_data
        if feature_type == "text":
            bins_for_hist = None

        return bins_for_hist

    def calculate_data_in_time(
        self,
        curr: pd.DataFrame,
        ref: Optional[pd.DataFrame],
        feature_name: str,
        feature_type: str,
        datetime_name: str,
        merge_small_cat: Optional[int] = MAX_CATEGORIES,
    ):
        result = None
        if feature_type == "cat" and merge_small_cat is not None:
            if ref is not None:
                ref = ref.copy()
            curr, ref = self._transform_cat_data(curr.copy(), ref, feature_name, merge_small_cat)

        freq = self._choose_agg_period(datetime_name, ref, curr)
        df_for_time_plot_curr = (
            curr.assign(period=lambda x: x[datetime_name].dt.to_period(freq=freq))
            .loc[:, ["period", feature_name]]
            .copy()
        )
        df_for_time_plot_ref = None
        if ref is not None:
            df_for_time_plot_ref = (
                ref.assign(period=lambda x: x[datetime_name].dt.to_period(freq=freq))
                .loc[:, ["period", feature_name]]
                .copy()
            )
        if feature_type == "num":
            df_for_time_plot_curr = self._transform_df_to_time_mean_view(
                df_for_time_plot_curr,
                datetime_name,
                feature_name,
            )
            if df_for_time_plot_ref is not None:
                df_for_time_plot_ref = self._transform_df_to_time_mean_view(
                    df_for_time_plot_ref,
                    datetime_name,
                    feature_name,
                )
            result = {
                "current": df_for_time_plot_curr,
                "reference": df_for_time_plot_ref,
                "freq": self.period_prefix,
                "datetime_name": datetime_name,
            }

        if feature_type == "cat":
            df_for_time_plot_curr = self._transform_df_to_time_count_view(
                df_for_time_plot_curr,
                datetime_name,
                feature_name,
            )
            if df_for_time_plot_ref is not None:
                df_for_time_plot_ref = self._transform_df_to_time_count_view(
                    df_for_time_plot_ref,
                    datetime_name,
                    feature_name,
                )
            result = {
                "current": df_for_time_plot_curr,
                "reference": df_for_time_plot_ref,
                "freq": self.period_prefix,
                "datetime_name": datetime_name,
            }

        return result

    def calculate_data_by_target(
        self,
        curr: pd.DataFrame,
        ref: Optional[pd.DataFrame],
        feature_name: str,
        feature_type: str,
        target_name: str,
        target_type: str,
        merge_small_cat: Optional[int] = MAX_CATEGORIES,
    ):
        result = None
        if feature_type == "cat" and target_type == "num":
            if ref is not None:
                ref = ref.copy()
            if merge_small_cat is not None:
                curr, ref = self._transform_cat_data(curr.copy(), ref, feature_name, merge_small_cat)
            result = self._prepare_box_data(curr, ref, feature_name, target_name)
        if feature_type == "num" and target_type == "cat":
            if ref is not None:
                ref = ref.copy()
            if merge_small_cat is not None:
                curr, ref = self._transform_cat_data(curr.copy(), ref, target_name, merge_small_cat)
            result = self._prepare_box_data(curr, ref, target_name, feature_name)
        if feature_type == "num" and target_type == "num":
            result = {}
            result["current"] = {
                feature_name: curr[feature_name].tolist(),
                target_name: curr[target_name].tolist(),
            }
            if ref is not None:
                result["reference"] = {
                    feature_name: ref[feature_name].tolist(),
                    target_name: ref[target_name].tolist(),
                }
        if feature_type == "cat" and target_type == "cat":
            if ref is not None:
                ref = ref.copy()
            if merge_small_cat is not None:
                curr, ref = self._transform_cat_data(curr.copy(), ref, feature_name, merge_small_cat)
            if ref is not None:
                ref = ref.copy()
            if merge_small_cat is not None:
                curr, ref = self._transform_cat_data(curr.copy(), ref, target_name, merge_small_cat, True)
            result = {}
            result["current"] = self._get_count_values(curr, target_name, feature_name)
            if ref is not None:
                result["reference"] = self._get_count_values(ref, target_name, feature_name)

        return result

    def _split_periods(self, curr_data, ref_data, feature_name):
        max_ref_date = ref_data[feature_name].max()
        min_curr_date = curr_data[feature_name].min()

        if (
            curr_data.loc[curr_data[feature_name] == min_curr_date, "number_of_items"].iloc[0]
            > ref_data.loc[ref_data[feature_name] == max_ref_date, "number_of_items"].iloc[0]
        ):
            curr_data.loc[curr_data[feature_name] == min_curr_date, "number_of_items"] = (
                curr_data.loc[curr_data[feature_name] == min_curr_date, "number_of_items"]
                + ref_data.loc[ref_data[feature_name] == max_ref_date, "number_of_items"]
            )
            ref_data = ref_data[ref_data[feature_name] != max_ref_date]
        else:
            ref_data.loc[ref_data[feature_name] == max_ref_date, "number_of_items"] = (
                ref_data.loc[ref_data[feature_name] == max_ref_date, "number_of_items"]
                + curr_data.loc[curr_data[feature_name] == min_curr_date, "number_of_items"]
            )
            curr_data = curr_data[curr_data[feature_name] != min_curr_date]
        return curr_data, ref_data

    def _get_count_values(self, df: pd.DataFrame, target_column: str, feature_name: Optional[str]):
        df = df.groupby([target_column, feature_name]).size()
        df.name = "count_objects"
        df = df.reset_index()
        return df[df["count_objects"] > 0]

    def _prepare_box_data(
        self,
        curr: pd.DataFrame,
        ref: Optional[pd.DataFrame],
        cat_feature_name: str,
        num_feature_name: str,
    ) -> Dict[str, Dict[str, list]]:
        dfs = [curr]
        names = ["current"]
        if ref is not None:
            dfs.append(ref)
            names.append("reference")
        res = {}
        for df, name in zip(dfs, names):
            df_for_plot = df.groupby(cat_feature_name)[num_feature_name].quantile([0, 0.25, 0.5, 0.75, 1]).reset_index()
            df_for_plot.columns = [cat_feature_name, "q", num_feature_name]
            res_df = {}
            values = df_for_plot[cat_feature_name].unique()

            def _quantiles(qdf, value):
                return qdf[df_for_plot.q == value].set_index(cat_feature_name).loc[values, num_feature_name].tolist()

            res_df["mins"] = _quantiles(df_for_plot, 0)
            res_df["lowers"] = _quantiles(df_for_plot, 0.25)
            res_df["means"] = _quantiles(df_for_plot, 0.5)
            res_df["uppers"] = _quantiles(df_for_plot, 0.75)
            res_df["maxs"] = _quantiles(df_for_plot, 1)
            res_df["values"] = values
            res[name] = res_df
        return res

    def _transform_cat_data(
        self,
        curr: pd.DataFrame,
        ref: Optional[pd.DataFrame],
        feature_name: str,
        merge_small_cat: int,
        rewrite: bool = False,
    ) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
        if self.curr is not None and rewrite is not True:
            return self.curr, self.ref
        if ref is not None:
            unique_values = len(
                np.union1d(curr[feature_name].astype(str).unique(), ref[feature_name].astype(str).unique())
            )
        else:
            unique_values = curr[feature_name].astype(str).nunique()

        if unique_values > merge_small_cat:
            curr_cats = curr[feature_name].astype(str).value_counts(normalize=True)

            if ref is not None:
                ref_cats = ref[feature_name].astype(str).value_counts(normalize=True)
                categories = pd.concat([curr_cats, ref_cats])

            else:
                categories = curr_cats

            cats = categories.sort_values(ascending=False).index.drop_duplicates(keep="first")[:merge_small_cat].values

            curr[feature_name] = curr[feature_name].apply(lambda x: x if str(x) in cats else "other")
            if ref is not None:
                ref[feature_name] = ref[feature_name].apply(lambda x: x if str(x) in cats else "other")
            self.curr = curr
            self.ref = ref
        return curr, ref

    def _choose_agg_period(
        self,
        date_column: str,
        reference_data: Optional[pd.DataFrame],
        current_data: pd.DataFrame,
    ) -> str:
        optimal_points = 150
        prefix_dict = {"A": "year", "Q": "quarter", "M": "month", "W": "week", "D": "day", "H": "hour"}
        datetime_feature = current_data[date_column]
        if reference_data is not None:
            datetime_feature = datetime_feature.append(reference_data[date_column])
        days = (datetime_feature.max() - datetime_feature.min()).days
        time_points = pd.Series(
            index=["A", "Q", "M", "W", "D", "H"],
            data=[
                abs(optimal_points - days / 365),
                abs(optimal_points - days / 90),
                abs(optimal_points - days / 30),
                abs(optimal_points - days / 7),
                abs(optimal_points - days),
                abs(optimal_points - days * 24),
            ],
        )
        self.period_prefix = prefix_dict[time_points.idxmin()]
        return str(time_points.idxmin())

    def _transform_df_to_time_mean_view(self, df: pd.DataFrame, date_column: str, feature_name: str):
        df = df.groupby("period")[feature_name].mean().reset_index()
        df[date_column] = df["period"].dt.to_timestamp()
        return df

    def _transform_df_to_time_count_view(self, df: pd.DataFrame, date_column: str, feature_name: str):
        df = df.groupby(["period", feature_name]).size()
        df.name = "num"
        df = df.reset_index()
        df[date_column] = df["period"].dt.to_timestamp()
        return df[df["num"] > 0]


def _select_features_for_corr(dataset: pd.DataFrame, columns: DatasetColumns) -> tuple:
    """Define which features should be used for calculating correlation matrices:
        - for pearson, spearman, and kendall correlation matrices we select numerical features which have > 1
            unique values;
        - for kramer_v correlation matrix, we select categorical features which have > 1 unique values.
    Args:
        columns: all columns data information.
    Returns:
        num_for_corr: list of feature names for pearson, spearman, and kendall correlation matrices.
        cat_for_corr: list of feature names for kramer_v correlation matrix.
    """
    target_name = columns.utility_columns.target
    prediction_name = columns.utility_columns.prediction
    num_for_corr = []
    for feature in columns.num_feature_names:
        unique_count = dataset[feature].nunique()

        if unique_count and unique_count > 1:
            num_for_corr.append(feature)

    cat_for_corr = []

    for feature in columns.cat_feature_names:
        unique_count = dataset[feature].nunique()

        if unique_count and unique_count > 1:
            cat_for_corr.append(feature)

    if target_name is not None:
        unique_count = dataset[target_name].nunique()

        if unique_count > 1:
            if columns.task == "classification":
                cat_for_corr.append(target_name)

            else:
                num_for_corr.append(target_name)

    if isinstance(prediction_name, str):
        unique_count = dataset[prediction_name].nunique()

        if unique_count > 1:
            if columns.task == "classification":
                cat_for_corr.append(prediction_name)

            else:
                num_for_corr.append(prediction_name)

    return num_for_corr, cat_for_corr


def _cramer_v(x: pd.Series, y: pd.Series) -> float:
    """Calculate Cramér's V: a measure of association between two nominal variables.
    Args:
        x: The array of observed values.
        y: The array of observed values.
    Returns:
        Value of the Cramér's V
    """
    arr = pd.crosstab(x, y).values
    chi2_stat = chi2_contingency(arr, correction=False)
    phi2 = chi2_stat[0] / arr.sum()
    n_rows, n_cols = arr.shape
    if min(n_cols - 1, n_rows - 1) == 0:
        value = np.nan
    else:
        value = np.sqrt(phi2 / min(n_cols - 1, n_rows - 1))

    return value


def get_pairwise_correlation(df, func: Callable[[pd.Series, pd.Series], float]) -> pd.DataFrame:
    """Compute pairwise correlation of columns
    Args:
        df: initial data frame.
        func: function for computing pairwise correlation.
    Returns:
        Correlation matrix.
    """
    columns = df.columns
    k = df.shape[1]
    if k <= 1:
        return pd.DataFrame()
    else:
        corr_array = np.eye(k)

        for i in range(k):
            for j in range(k):
                if i <= j:
                    continue
                c = func(df[columns[i]], df[columns[j]])
                corr_array[i, j] = c
                corr_array[j, i] = c
        return pd.DataFrame(data=corr_array, columns=columns, index=columns)


def _calculate_correlations(df: pd.DataFrame, num_for_corr, cat_for_corr, kind):
    """Calculate correlation matrix depending on the kind parameter
    Args:
        df: initial data frame.
        num_for_corr: list of feature names for pearson, spearman, and kendall correlation matrices.
        cat_for_corr: list of feature names for kramer_v correlation matrix.
        kind: Method of correlation:
            - pearson - standard correlation coefficient
            - kendall - Kendall Tau correlation coefficient
            - spearman - Spearman rank correlation
            - cramer_v - Cramer’s V measure of association
    Returns:
        Correlation matrix.
    """
    if kind == "pearson":
        return df[num_for_corr].corr("pearson")
    elif kind == "spearman":
        return df[num_for_corr].corr("spearman")
    elif kind == "kendall":
        return df[num_for_corr].corr("kendall")
    elif kind == "cramer_v":
        return get_pairwise_correlation(df[cat_for_corr], _cramer_v)


def calculate_correlations(
    dataset: pd.DataFrame,
    columns: DatasetColumns,
    add_text_columns: Optional[list] = None,
) -> Dict:
    num_for_corr, cat_for_corr = _select_features_for_corr(dataset, columns)
    if add_text_columns is not None:
        num_for_corr += add_text_columns
    correlations = {}

    for kind in ["pearson", "spearman", "kendall", "cramer_v"]:
        correlations[kind] = _calculate_correlations(dataset, num_for_corr, cat_for_corr, kind)

    return correlations


@dataclasses.dataclass
class ColumnCorrelations:
    column_name: str
    kind: str
    values: Distribution


def calculate_cramer_v_correlation(column_name: str, dataset: pd.DataFrame, columns: List[str]) -> ColumnCorrelations:
    result_x = []
    result_y = []

    if not dataset[column_name].empty:
        for correlation_columns_name in columns:
            result_x.append(correlation_columns_name)
            result_y.append(_cramer_v(dataset[column_name], dataset[correlation_columns_name]))

    return ColumnCorrelations(column_name=column_name, kind="cramer_v", values=Distribution(x=result_x, y=result_y))


def calculate_category_column_correlations(
    column_name: str, dataset: pd.DataFrame, columns: List[str]
) -> Dict[str, ColumnCorrelations]:
    """For category columns calculate cramer_v correlation"""
    if dataset[column_name].empty:
        return {}

    if not columns:
        return {}

    correlation = calculate_cramer_v_correlation(column_name, dataset, columns)

    if pd.isnull(correlation.values.y).all():
        return {}

    else:
        return {correlation.kind: correlation}


def calculate_numerical_column_correlations(
    column_name: str, dataset: pd.DataFrame, columns: List[str]
) -> Dict[str, ColumnCorrelations]:

    if dataset[column_name].empty or not columns:
        return {}

    if not columns:
        return {}

    result: Dict[str, ColumnCorrelations] = {}
    column = dataset[column_name]

    for kind in ["pearson", "spearman", "kendall"]:
        correlations_columns = []
        correlations_values = []

        for other_column_name in columns:
            correlations_columns.append(other_column_name)
            correlations_values.append(column.corr(dataset[other_column_name], method=kind))

        result[kind] = ColumnCorrelations(
            column_name=column_name,
            kind=kind,
            values=Distribution(x=correlations_columns, y=correlations_values),
        )

    return result


def calculate_column_distribution(column: pd.Series, column_type: str) -> ColumnDistribution:
    if column.empty:
        distribution: ColumnDistribution = {}

    elif column_type == "num":
        # TODO: implement distribution for num column
        value_counts = column.value_counts(dropna=True)
        distribution = dict(value_counts)

    elif column_type == "cat":
        value_counts = column.value_counts(dropna=True)
        distribution = dict(value_counts)

    else:
        raise ValueError(f"Cannot calculate distribution for column type {column_type}")

    return distribution



import dataclasses
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import numpy as np
import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import table_data
from evidently.renderers.html_widgets import widget_tabs


@dataclasses.dataclass
class ColumnMissingValues:
    """Statistics about missing values in a column"""

    # count of rows in the column
    number_of_rows: int
    # set of different missed values in the column
    different_missing_values: Dict[Any, int]
    # number of different missed values in the column
    number_of_different_missing_values: int
    # count of missed values in the column
    number_of_missing_values: int
    # share of missed values in the column
    share_of_missing_values: float


@dataclasses.dataclass
class ColumnMissingValuesMetricResult:
    column_name: str
    current: ColumnMissingValues
    reference: Optional[ColumnMissingValues] = None


class ColumnMissingValuesMetric(Metric[ColumnMissingValuesMetricResult]):
    """Count missing values in a column.

    Missing value is a null or NaN value.

    Calculate an amount of missing values kinds and count for such values.
    NA-types like numpy.NaN, pandas.NaT are counted as one type.

    You can set you own missing values list with `missing_values` parameter.
    Value None in the list means that Pandas null values will be included in the calculation.

    If `replace` parameter is False - add defaults to user's list.
    If `replace` parameter is True - use values from `missing_values` list only.
    """

    # default missing values list
    DEFAULT_MISSING_VALUES = ["", np.inf, -np.inf, None]
    missing_values: frozenset
    column_name: str

    def __init__(self, column_name: str, missing_values: Optional[list] = None, replace: bool = True) -> None:
        self.column_name = column_name

        if missing_values is None:
            # use default missing values list if we have no user-defined missed values
            missing_values = self.DEFAULT_MISSING_VALUES

        elif not replace:
            # add default missing values to user-defined list
            missing_values = self.DEFAULT_MISSING_VALUES + missing_values

        # use frozenset because metrics parameters should be immutable/hashable for deduplication
        self.missing_values = frozenset(missing_values)

    def _calculate_missing_values_stats(self, column: pd.Series) -> ColumnMissingValues:
        different_missing_values = {value: 0 for value in self.missing_values}
        number_of_missing_values = 0

        number_of_rows = len(column)

        # iterate by each value in custom missing values list and check the value in a column
        for value in self.missing_values:
            if value is None:
                # check all pandas missing values like numpy.NAN, pandas.NA, pandas.NaT, etc
                missing_values = column.isnull().sum()

            else:
                missing_values = (column == value).sum()

            if missing_values > 0:
                # increase overall counter
                number_of_missing_values += missing_values
                # increase by-missing-value counter
                different_missing_values[value] += missing_values

        share_of_missing_values = number_of_missing_values / number_of_rows

        # sort by missing values count
        different_missing_values = {
            value: count
            for value, count in sorted(different_missing_values.items(), key=lambda item: item[1], reverse=True)
        }

        number_of_different_missing_values = sum(
            [1 for value in different_missing_values if different_missing_values[value] > 0]
        )

        return ColumnMissingValues(
            different_missing_values=different_missing_values,
            number_of_different_missing_values=number_of_different_missing_values,
            number_of_missing_values=number_of_missing_values,
            share_of_missing_values=share_of_missing_values,
            number_of_rows=number_of_rows,
        )

    def calculate(self, data: InputData) -> ColumnMissingValuesMetricResult:
        if not self.missing_values:
            raise ValueError("Missed values list should not be empty.")

        if self.column_name not in data.current_data:
            raise ValueError(f"Column {self.column_name} is not in current data.")

        current_missing_values = self._calculate_missing_values_stats(data.current_data[self.column_name])

        if data.reference_data is None:
            reference_missing_values: Optional[ColumnMissingValues] = None

        else:
            if self.column_name not in data.reference_data:
                raise ValueError(f"Column {self.column_name} is not in reference data.")

            reference_missing_values = self._calculate_missing_values_stats(data.reference_data[self.column_name])

        return ColumnMissingValuesMetricResult(
            column_name=self.column_name,
            current=current_missing_values,
            reference=reference_missing_values,
        )


@default_renderer(wrap_type=ColumnMissingValuesMetric)
class ColumnMissingValuesMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnMissingValuesMetric) -> dict:
        return dataclasses.asdict(obj.get_result())

    @staticmethod
    def _get_table_stat(stats: ColumnMissingValues) -> BaseWidgetInfo:
        data = []

        for missed_value, count_of_missed in stats.different_missing_values.items():
            percent_of_missed = round(count_of_missed / stats.number_of_rows * 100, 3)
            value = f"{count_of_missed} ({percent_of_missed}%)"

            if missed_value is None:
                missed_value_str = "Pandas and Numpy NA, NaN, etc."

            elif not missed_value:
                missed_value_str = "Empty string"

            else:
                missed_value_str = str(missed_value)

            data.append((missed_value_str, value))

        matched_stat_headers = ["Missing values type", "Count"]
        return table_data(
            title="",
            column_names=matched_stat_headers,
            data=data,
        )

    @staticmethod
    def _get_info_string(stats: ColumnMissingValues) -> str:
        percents = round(stats.share_of_missing_values * 100, 3)
        return f"{stats.number_of_missing_values} ({percents}%)"

    def _get_details_missing_values_info(self, metric_result: ColumnMissingValuesMetricResult) -> BaseWidgetInfo:
        counters = [
            CounterData.string("Missing values (Current data)", self._get_info_string(metric_result.current)),
        ]
        if metric_result.reference is not None:
            counters.append(
                CounterData.string("Missing values (Reference data)", self._get_info_string(metric_result.reference)),
            )

        return counter(
            title="",
            counters=counters,
        )

    def render_html(self, obj: ColumnMissingValuesMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()

        result = [
            header_text(label=f"Missing values in column '{metric_result.column_name}'"),
            self._get_details_missing_values_info(metric_result=metric_result),
        ]

        current_table = self._get_table_stat(metric_result.current)

        if metric_result.reference is not None:
            tables = widget_tabs(
                tabs=[
                    TabData(title="Current dataset", widget=current_table),
                    TabData(
                        title="Reference dataset",
                        widget=self._get_table_stat(metric_result.reference),
                    ),
                ]
            )

        else:
            tables = current_table

        result.append(tables)
        return result



from itertools import combinations

import pandas as pd


def get_number_of_all_pandas_missed_values(dataset: pd.DataFrame) -> int:
    """Calculate the number of missed - nulls by pandas - values in a dataset"""
    return dataset.isnull().sum().sum()


def get_number_of_empty_columns(dataset: pd.DataFrame) -> int:
    """Calculate the number of empty columns in a dataset"""
    return dataset.isnull().all().sum()


def get_number_of_duplicated_columns(dataset: pd.DataFrame) -> int:
    """Calculate the number of duplicated columns in a dataset"""
    return sum([1 for i, j in combinations(dataset, 2) if dataset[i].equals(dataset[j])])


def get_number_of_almost_duplicated_columns(dataset: pd.DataFrame, threshold: float) -> int:
    """Calculate the number of almost duplicated columns in a dataset"""
    result = 0
    all_rows = dataset.shape[0]
    checked_columns = set()

    for column_name_1 in dataset.columns:
        checked_columns.add(column_name_1)

        for column_name_2 in dataset.columns:
            if column_name_2 in checked_columns:
                continue

            # if dtypes of columns are different, then columns are not duplicated
            # check names because of problems with categorical columns
            if dataset[column_name_1].dtype.name != dataset[column_name_2].dtype.name:
                continue

            # if columns are categorical, then we need to check categories lists
            # if the lists are not the same, Series.eq method raises an exception
            if pd.api.types.is_categorical_dtype(dataset[column_name_1]) and pd.api.types.is_categorical_dtype(
                dataset[column_name_2]
            ):
                if dataset[column_name_1].cat.categories.tolist() != dataset[column_name_2].cat.categories.tolist():
                    continue

            score = dataset[column_name_1].eq(dataset[column_name_2]).sum() / all_rows

            if score >= threshold:
                result += 1

    return result


def get_number_of_constant_columns(dataset: pd.DataFrame) -> int:
    """Calculate the number of constant columns in a dataset"""
    return len(dataset.columns[dataset.nunique() <= 1])  # type: ignore


def get_number_of_almost_constant_columns(dataset: pd.DataFrame, threshold: float) -> int:
    """Calculate the number of almost constant columns in a dataset"""
    result = 0
    dataset_row = dataset.shape[0]

    if dataset_row <= 0:
        return 0

    for column_name in dataset.columns:
        score = dataset[column_name].value_counts().max() / dataset.shape[0]

        if score >= threshold:
            result += 1

    return result



"""Methods and types for data drift calculations."""

from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
import pandas as pd

from evidently.calculations.stattests import get_stattest
from evidently.options import DataDriftOptions
from evidently.utils.data_drift_utils import get_text_data_for_plots
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import recognize_column_type
from evidently.utils.visualizations import Distribution
from evidently.utils.visualizations import get_distribution_for_column


@dataclass
class ColumnDataDriftMetrics:
    """One column drift metrics."""

    column_name: str
    column_type: str
    stattest_name: str
    drift_score: float
    drift_detected: bool
    threshold: float
    # distributions for the column
    current_distribution: Optional[Distribution]
    reference_distribution: Optional[Distribution]
    current_small_distribution: Optional[list] = None
    reference_small_distribution: Optional[list] = None
    # data for scatter plot for numeric features only
    current_scatter: Optional[Dict[str, list]] = None
    x_name: Optional[str] = None
    plot_shape: Optional[Dict[str, float]] = None
    # correlations for numeric features only
    current_correlations: Optional[Dict[str, float]] = None
    reference_correlations: Optional[Dict[str, float]] = None
    typical_examples_cur: Optional[List[str]] = None
    typical_examples_ref: Optional[List[str]] = None
    typical_words_cur: Optional[List[str]] = None
    typical_words_ref: Optional[List[str]] = None


@dataclass
class DatasetDrift:
    """Dataset drift calculation results"""

    number_of_drifted_columns: int
    dataset_drift_score: float
    dataset_drift: bool


@dataclass
class DatasetDriftMetrics:
    number_of_columns: int
    number_of_drifted_columns: int
    share_of_drifted_columns: float
    dataset_drift: bool
    drift_by_columns: Dict[str, ColumnDataDriftMetrics]
    options: DataDriftOptions
    dataset_columns: DatasetColumns


def get_one_column_drift(
    *,
    current_data: pd.DataFrame,
    reference_data: pd.DataFrame,
    column_name: str,
    options: DataDriftOptions,
    dataset_columns: DatasetColumns,
    column_type: Optional[str] = None,
) -> ColumnDataDriftMetrics:
    if column_name not in current_data:
        raise ValueError(f"Cannot find column '{column_name}' in current dataset")

    if column_name not in reference_data:
        raise ValueError(f"Cannot find column '{column_name}' in reference dataset")

    if column_type is None:
        column_type = recognize_column_type(dataset=reference_data, column_name=column_name, columns=dataset_columns)

    if column_type not in ("cat", "num", "text"):
        raise ValueError(f"Cannot calculate drift metric for column '{column_name}' with type {column_type}")

    stattest = None

    if column_name == dataset_columns.utility_columns.target and column_type == "num":
        stattest = options.num_target_stattest_func

    elif column_name == dataset_columns.utility_columns.target and column_type == "cat":
        stattest = options.cat_target_stattest_func

    if not stattest:
        stattest = options.get_feature_stattest_func(column_name, column_type)

    threshold = options.get_threshold(column_name, column_type)
    current_column = current_data[column_name]
    reference_column = reference_data[column_name]

    # clean and check the column in reference dataset
    reference_column = reference_column.replace([-np.inf, np.inf], np.nan).dropna()

    if reference_column.empty:
        raise ValueError(
            f"An empty column '{column_name}' was provided for drift calculation in the reference dataset."
        )

    # clean and check the column in current dataset
    current_column = current_column.replace([-np.inf, np.inf], np.nan).dropna()

    if current_column.empty:
        raise ValueError(f"An empty column '{column_name}' was provided for drift calculation in the current dataset.")

    current_distribution = None
    reference_distribution = None
    current_small_distribution = None
    reference_small_distribution = None
    current_correlations = None
    reference_correlations = None
    current_scatter = None
    x_name = None
    plot_shape = None
    typical_examples_cur = None
    typical_examples_ref = None
    typical_words_cur = None
    typical_words_ref = None

    if column_type == "num":
        if not pd.api.types.is_numeric_dtype(reference_column):
            raise ValueError(f"Column '{column_name}' in reference dataset should contain numerical values only.")

        if not pd.api.types.is_numeric_dtype(current_column):
            raise ValueError(f"Column '{column_name}' in current dataset should contain numerical values only.")

    drift_test_function = get_stattest(reference_column, current_column, column_type, stattest)
    drift_result = drift_test_function(reference_column, current_column, column_type, threshold)

    if column_type == "num":
        numeric_columns = dataset_columns.num_feature_names

        if column_name not in numeric_columns:
            # for target and prediction cases add the column_name in the numeric columns list
            numeric_columns = numeric_columns + [column_name]

        current_correlations = current_data[numeric_columns].corr()[column_name].to_dict()
        reference_correlations = reference_data[numeric_columns].corr()[column_name].to_dict()
        current_nbinsx = options.get_nbinsx(column_name)
        current_small_distribution = [
            t.tolist()
            for t in np.histogram(
                current_data[column_name][np.isfinite(current_data[column_name])],
                bins=current_nbinsx,
                density=True,
            )
        ]
        reference_small_distribution = [
            t.tolist()
            for t in np.histogram(
                reference_data[column_name][np.isfinite(reference_data[column_name])],
                bins=current_nbinsx,
                density=True,
            )
        ]
        current_scatter = {}
        current_scatter[column_name] = current_data[column_name]
        datetime_column_name = dataset_columns.utility_columns.date
        if datetime_column_name is not None:
            current_scatter["Timestamp"] = current_data[datetime_column_name]
            x_name = "Timestamp"
        else:
            current_scatter["Index"] = current_data.index
            x_name = "Index"

        plot_shape = {}
        reference_mean = reference_data[column_name].mean()
        reference_std = reference_data[column_name].std()
        plot_shape["y0"] = reference_mean - reference_std
        plot_shape["y1"] = reference_mean + reference_std

    elif column_type == "cat":
        reference_counts = reference_data[column_name].value_counts(sort=False)
        current_counts = current_data[column_name].value_counts(sort=False)
        keys = set(reference_counts.keys()).union(set(current_counts.keys()))

        for key in keys:
            if key not in reference_counts:
                reference_counts.loc[key] = 0
            if key not in current_counts:
                current_counts.loc[key] = 0

        reference_small_distribution = list(
            reversed(list(map(list, zip(*sorted(reference_counts.items(), key=lambda x: str(x[0]))))))
        )
        current_small_distribution = list(
            reversed(list(map(list, zip(*sorted(current_counts.items(), key=lambda x: str(x[0]))))))
        )
    if column_type != "text":
        current_distribution, reference_distribution = get_distribution_for_column(
            column_type=column_type,
            current=current_column,
            reference=reference_column,
        )
        if reference_distribution is None:
            raise ValueError(f"Cannot calculate reference distribution for column '{column_name}'.")

    elif column_type == "text" and drift_result.drifted:
        typical_examples_cur, typical_examples_ref, typical_words_cur, typical_words_ref = get_text_data_for_plots(
            reference_column, current_column
        )

    return ColumnDataDriftMetrics(
        column_name=column_name,
        column_type=column_type,
        stattest_name=drift_test_function.display_name,
        drift_score=drift_result.drift_score,
        drift_detected=drift_result.drifted,
        threshold=drift_result.actual_threshold,
        current_distribution=current_distribution,
        reference_distribution=reference_distribution,
        current_small_distribution=current_small_distribution,
        reference_small_distribution=reference_small_distribution,
        current_correlations=current_correlations,
        reference_correlations=reference_correlations,
        current_scatter=current_scatter,
        x_name=x_name,
        plot_shape=plot_shape,
        typical_examples_cur=typical_examples_cur,
        typical_examples_ref=typical_examples_ref,
        typical_words_cur=typical_words_cur,
        typical_words_ref=typical_words_ref,
    )


def _get_pred_labels_from_prob(dataframe: pd.DataFrame, prediction_column: list) -> List[str]:
    """Get labels from probabilities from columns by prediction columns list"""
    array_prediction = dataframe[prediction_column].to_numpy()
    prediction_ids = np.argmax(array_prediction, axis=-1)
    prediction_labels = [prediction_column[x] for x in prediction_ids]
    return prediction_labels


def ensure_prediction_column_is_string(
    *,
    prediction_column: Optional[Union[str, Sequence]],
    current_data: pd.DataFrame,
    reference_data: pd.DataFrame,
    threshold: float = 0.5,
) -> Optional[str]:
    """Update dataset by predictions type:

    - if prediction column is None or a string, no dataset changes
    - (binary classification) if predictions is a list and its length equals 2
        set predicted_labels column by `threshold`
    - (multy label classification) if predictions is a list and its length is greater than 2
        set predicted_labels from probability values in columns by prediction column


    Returns:
         prediction column name.
    """
    result_prediction_column = None

    if prediction_column is None or isinstance(prediction_column, str):
        result_prediction_column = prediction_column

    elif isinstance(prediction_column, list):
        if len(prediction_column) > 2:
            reference_data["predicted_labels"] = _get_pred_labels_from_prob(reference_data, prediction_column)
            current_data["predicted_labels"] = _get_pred_labels_from_prob(current_data, prediction_column)
            result_prediction_column = "predicted_labels"

        elif len(prediction_column) == 2:
            reference_data["predicted_labels"] = (reference_data[prediction_column[0]] > threshold).astype(int)
            current_data["predicted_labels"] = (current_data[prediction_column[0]] > threshold).astype(int)
            result_prediction_column = "predicted_labels"

    return result_prediction_column


def get_dataset_drift(drift_metrics, drift_share=0.5) -> DatasetDrift:
    number_of_drifted_columns = sum([1 if drift.drift_detected else 0 for _, drift in drift_metrics.items()])
    share_drifted_columns = number_of_drifted_columns / len(drift_metrics)
    dataset_drift = bool(share_drifted_columns >= drift_share)
    return DatasetDrift(
        number_of_drifted_columns=number_of_drifted_columns,
        dataset_drift_score=share_drifted_columns,
        dataset_drift=dataset_drift,
    )


def _get_all_columns_for_drift(dataset_columns: DatasetColumns) -> List[str]:
    result = []
    target_column = dataset_columns.utility_columns.target

    if target_column:
        result.append(target_column)

    prediction_column = dataset_columns.utility_columns.prediction

    if isinstance(prediction_column, str):
        result.append(prediction_column)

    if dataset_columns.num_feature_names:
        result += dataset_columns.num_feature_names

    if dataset_columns.cat_feature_names:
        result += dataset_columns.cat_feature_names

    if dataset_columns.text_feature_names:
        result += dataset_columns.text_feature_names

    return result


def get_drift_for_columns(
    *,
    current_data: pd.DataFrame,
    reference_data: pd.DataFrame,
    dataset_columns: DatasetColumns,
    data_drift_options: DataDriftOptions,
    drift_share_threshold: Optional[float] = None,
    columns: Optional[List[str]] = None,
) -> DatasetDriftMetrics:
    if columns is None:
        # ensure prediction column is a string - add label values for classification tasks
        ensure_prediction_column_is_string(
            prediction_column=dataset_columns.utility_columns.prediction,
            current_data=current_data,
            reference_data=reference_data,
        )
        columns = _get_all_columns_for_drift(dataset_columns)

    drift_share_threshold = drift_share_threshold or data_drift_options.drift_share

    # calculate result
    drift_by_columns = {}

    for column_name in columns:
        drift_by_columns[column_name] = get_one_column_drift(
            current_data=current_data,
            reference_data=reference_data,
            column_name=column_name,
            options=data_drift_options,
            dataset_columns=dataset_columns,
        )

    dataset_drift = get_dataset_drift(drift_by_columns, drift_share_threshold)
    return DatasetDriftMetrics(
        number_of_columns=len(columns),
        number_of_drifted_columns=dataset_drift.number_of_drifted_columns,
        share_of_drifted_columns=dataset_drift.dataset_drift_score,
        dataset_drift=dataset_drift.dataset_drift,
        drift_by_columns=drift_by_columns,
        options=data_drift_options,
        dataset_columns=dataset_columns,
    )



import dataclasses
from typing import List
from typing import Optional
from typing import Tuple

import numpy as np
import pandas as pd
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.linear_model import SGDClassifier
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import table_data
from evidently.renderers.html_widgets import widget_tabs


@dataclasses.dataclass
class TextDomainClassifierDriftResult:
    text_column_name: str
    domain_classifier_roc_auc: float
    random_classifier_95_percentile: float
    content_drift: bool
    characteristic_examples_current: Optional[List[str]]
    characteristic_examples_reference: Optional[List[str]]
    characteristic_words_current: Optional[List[str]]
    characteristic_words_reference: Optional[List[str]]


class TextDomainClassifierDriftMetric(Metric[TextDomainClassifierDriftResult]):
    text_column_name: str

    def __init__(self, text_column_name: str) -> None:
        self.text_column_name = text_column_name

    @staticmethod
    def roc_auc_domain_classifier(X_train, X_test, y_train, y_test) -> Tuple:
        pipeline = Pipeline(
            [
                ("vectorization", TfidfVectorizer(sublinear_tf=True, max_df=0.5, stop_words="english")),
                ("classification", SGDClassifier(alpha=0.0001, max_iter=50, penalty="l1", loss="modified_huber")),
            ]
        )

        pipeline.fit(X_train, y_train)
        y_pred_proba = pipeline.predict_proba(X_test)[:, 1]
        roc_auc = roc_auc_score(y_test, y_pred_proba)
        return roc_auc, y_pred_proba, pipeline

    @staticmethod
    def roc_auc_random_classifier_percentile(y_test: np.array, p_value=0.05, iter_num=1000, seed=42) -> float:
        def calc_roc_auc_random(y_test, seed=None):
            np.random.seed(seed)
            y_random_pred = np.random.rand(len(y_test))
            roc_auc_random = roc_auc_score(y_test, y_random_pred)
            return roc_auc_random

        np.random.seed(seed)
        seeds = np.random.randint(0, iter_num * 10, size=iter_num)
        roc_auc_values = [calc_roc_auc_random(y_test, seed=seed) for seed in seeds]
        return np.percentile(roc_auc_values, 100 * (1 - p_value))

    @staticmethod
    def is_drift_detected(roc_auc_classifier_value: float, roc_auc_random_percentile: float) -> bool:
        if roc_auc_classifier_value > roc_auc_random_percentile:
            return True
        else:
            return False

    @staticmethod
    def get_typical_examples(X_test, y_test, y_pred_proba, examples_num=10) -> Tuple[List[str], List[str]]:
        typical_examples = pd.DataFrame({"text": X_test, "label": y_test, "predict_proba": y_pred_proba})

        typical_current = typical_examples[typical_examples["label"] == 1]
        typical_current.sort_values("predict_proba", ascending=False, inplace=True)
        typical_current = typical_current.head(examples_num)

        typical_reference = typical_examples[typical_examples["label"] == 0]
        typical_reference.sort_values("predict_proba", ascending=True, inplace=True)
        typical_reference = typical_reference.head(examples_num)
        return list(typical_current["text"]), list(typical_reference["text"])

    @staticmethod
    def get_typical_words(pipeline, words_num=25) -> Tuple[List[str], List[str]]:
        weights_df = pd.DataFrame({"weight": pipeline["classification"].coef_[0]})
        weights_df.reset_index(inplace=True)
        weights_df.rename(columns={"index": "feature_ind"}, inplace=True)
        weights_df = weights_df[weights_df["weight"] != 0]

        # build inverted index for vocabulary
        inverted_vocabulary = {value: key for key, value in pipeline["vectorization"].vocabulary_.items()}

        words_typical_current = weights_df[weights_df["weight"] > 0]
        words_typical_current.sort_values("weight", ascending=False, inplace=True)
        words_typical_current = words_typical_current.head(words_num)
        words_typical_current["word"] = words_typical_current["feature_ind"].apply(lambda x: inverted_vocabulary[x])

        words_typical_reference = weights_df[weights_df["weight"] < 0]
        words_typical_reference.sort_values("weight", ascending=True, inplace=True)
        words_typical_reference = words_typical_reference.head(words_num)
        words_typical_reference["word"] = words_typical_reference["feature_ind"].apply(lambda x: inverted_vocabulary[x])

        return list(words_typical_current["word"]), list(words_typical_reference["word"])

    def calculate(self, data: InputData) -> TextDomainClassifierDriftResult:
        # create datasets for domain classifier training and testing
        ref_data = pd.Series() if data.reference_data is None else data.reference_data[self.text_column_name]
        domain_data = pd.concat(
            [
                pd.DataFrame({"text": ref_data, "target": 0}),
                pd.DataFrame({"text": data.current_data[self.text_column_name], "target": 1}),
            ]
        )

        X_train, X_test, y_train, y_test = train_test_split(
            domain_data["text"],
            domain_data["target"],
            test_size=0.5,
            random_state=42,
            shuffle=True,
        )
        # calculate domain classifier roc-auc score
        domain_classifier_roc_auc, y_pred_proba, classifier_pipeline = self.roc_auc_domain_classifier(
            X_train,
            X_test,
            y_train,
            y_test,
        )

        # calculate roc-auc scores 95 percentile if labels were assigned randomly
        random_classifier_95_percentile = self.roc_auc_random_classifier_percentile(y_test)

        # detect data drift if present
        is_content_drift = self.is_drift_detected(
            roc_auc_classifier_value=domain_classifier_roc_auc,
            roc_auc_random_percentile=random_classifier_95_percentile,
        )

        typical_examples_cur = None
        typical_examples_ref = None
        typical_words_cur = None
        typical_words_ref = None

        if is_content_drift:
            # get examples more characteristic of current or reference dataset
            typical_examples_cur, typical_examples_ref = self.get_typical_examples(X_test, y_test, y_pred_proba)

            # get words more characteristic of current or reference dataset
            typical_words_cur, typical_words_ref = self.get_typical_words(classifier_pipeline)

        return TextDomainClassifierDriftResult(
            text_column_name=self.text_column_name,
            domain_classifier_roc_auc=domain_classifier_roc_auc,
            random_classifier_95_percentile=random_classifier_95_percentile,
            content_drift=is_content_drift,
            characteristic_examples_current=typical_examples_cur,
            characteristic_examples_reference=typical_examples_ref,
            characteristic_words_current=typical_words_cur,
            characteristic_words_reference=typical_words_ref,
        )


@default_renderer(wrap_type=TextDomainClassifierDriftMetric)
class TextDomainClassifierDriftMetricRenderer(MetricRenderer):
    def render_json(self, obj: TextDomainClassifierDriftMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        return result

    @staticmethod
    def _get_table_stat(dataset_name: str, metric_result, content_type="words") -> BaseWidgetInfo:
        if content_type == "words":
            if dataset_name == "current":
                data = metric_result.characteristic_words_current
            else:
                data = metric_result.characteristic_words_reference
        elif content_type == "examples":
            if dataset_name == "current":
                data = metric_result.characteristic_examples_current
            else:
                data = metric_result.characteristic_examples_reference
        else:
            raise Exception("Unknown content type {}. Supported types are ['words', 'examples']".format(content_type))
        res = table_data(
            title="{} Dataset: characteristic {}".format(dataset_name.capitalize(), content_type.capitalize()),
            column_names=["", ""],
            data=[[el, ""] for el in data],
        )
        return res

    def render_html(self, obj: TextDomainClassifierDriftMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        if metric_result.content_drift:
            drift_result_str = "DETECTED"
        else:
            drift_result_str = "NOT detected"

        result = [
            counter(
                counters=[
                    CounterData(
                        "Content drift is {}".format(drift_result_str),
                        "Content Drift for '{}'".format(metric_result.text_column_name),
                    )
                ],
                title="",
            )
        ]

        counters = [
            CounterData.float("Domain classifier ROC-AUC score", metric_result.domain_classifier_roc_auc, 3),
            CounterData.float(
                "95th percentile for random predictions",
                metric_result.random_classifier_95_percentile,
                3,
            ),
        ]

        result.append(counter(counters=counters, title=""))

        if metric_result.content_drift:
            current_table_words = self._get_table_stat("current", metric_result, content_type="words")
            reference_table_words = self._get_table_stat("reference", metric_result, content_type="words")

            current_table_examples = self._get_table_stat("current", metric_result, content_type="examples")
            reference_table_examples = self._get_table_stat("reference", metric_result, content_type="examples")

            tables_tabs = [
                TabData(title="Current dataset", widget=current_table_words),
                TabData(title="Reference dataset", widget=reference_table_words),
            ]
            tables = widget_tabs(tabs=tables_tabs)
            result.append(tables)

            tables_tabs = [
                TabData(title="Current dataset", widget=current_table_examples),
                TabData(title="Reference dataset", widget=reference_table_examples),
            ]
            tables = widget_tabs(tabs=tables_tabs)
            result.append(tables)
        return result



from typing import Optional
from typing import Union

from evidently.base_metric import InputData
from evidently.calculations.stattests import PossibleStatTestType
from evidently.test_preset.test_preset import TestPreset
from evidently.tests import TestAccuracyScore
from evidently.tests import TestColumnDrift
from evidently.tests import TestF1Score
from evidently.tests import TestLogLoss
from evidently.tests import TestPrecisionScore
from evidently.tests import TestRecallScore
from evidently.tests import TestRocAuc
from evidently.utils.data_operations import DatasetColumns


class BinaryClassificationTopKTestPreset(TestPreset):
    """
    Binary Classification Tests for Top K threshold.
    Args:
        stattest: stattest for `TestColumnDrift`
        stattest_threshold: threshold for stattest

    Contains:
    - `TestColumnDrift` for target
    - `TestPrecisionScore`
    - `TestRecallScore`
    - `TestF1Score`
    - `TestAccuracyScore`
    """

    stattest: Optional[PossibleStatTestType]
    stattest_threshold: Optional[float]

    def __init__(
        self,
        k: Union[float, int],
        stattest: Optional[PossibleStatTestType] = None,
        stattest_threshold: Optional[float] = None,
    ):
        super().__init__()
        self.k = k
        self.stattest = stattest
        self.stattest_threshold = stattest_threshold

    def generate_tests(self, data: InputData, columns: DatasetColumns):
        target = columns.utility_columns.target
        if target is None:
            raise ValueError("Target column should be set in mapping and be present in data")
        return [
            TestAccuracyScore(k=self.k),
            TestPrecisionScore(k=self.k),
            TestRecallScore(k=self.k),
            TestF1Score(k=self.k),
            TestColumnDrift(column_name=target, stattest=self.stattest, stattest_threshold=self.stattest_threshold),
            TestRocAuc(),
            TestLogLoss(),
        ]



#!/usr/bin/env python
# coding: utf-8
from typing import Optional

import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.prob_classification_performance_analyzer import (
    ProbClassificationPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class ProbClassQualityMetricBarWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [ProbClassificationPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        results = ProbClassificationPerformanceAnalyzer.get_results(analyzers_results)
        utility_columns = results.columns.utility_columns

        if utility_columns.target is None or utility_columns.prediction is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )

            return None

        if self.dataset == "reference":
            metrics = results.reference_metrics

            if metrics is None:
                raise ValueError(
                    f"Widget [{self.title}] required 'reference' results from"
                    f" {ProbClassificationPerformanceAnalyzer.__name__} but no data found"
                )

        elif self.dataset == "current":
            metrics = results.current_metrics

        else:
            raise ValueError(
                f"Widget [{self.title}] required 'current' or 'reference' dataset value"
            )

        if metrics is None:
            return None

        return BaseWidgetInfo(
            title=self.title,
            type="counter",
            size=2,
            params={
                "counters": [
                    {"value": str(round(metrics.accuracy, 3)), "label": "Accuracy"},
                    {"value": str(round(metrics.precision, 3)), "label": "Precision"},
                    {"value": str(round(metrics.recall, 3)), "label": "Recall"},
                    {"value": str(round(metrics.f1, 3)), "label": "F1"},
                    {"value": str(round(metrics.roc_auc, 3)), "label": "ROC AUC"},
                    {"value": str(round(metrics.log_loss, 3)), "label": "LogLoss"},
                ]
            },
        )



import json
from dataclasses import dataclass
from typing import Dict

from evidently.model_profile import Profile
from evidently.model_profile.sections.cat_target_drift_profile_section import CatTargetDriftProfileSection
from evidently.model_profile.sections.classification_performance_profile_section import (
    ClassificationPerformanceProfileSection,
)
from evidently.model_profile.sections.data_drift_profile_section import DataDriftProfileSection
from evidently.model_profile.sections.data_quality_profile_section import DataQualityProfileSection
from evidently.model_profile.sections.num_target_drift_profile_section import NumTargetDriftProfileSection
from evidently.model_profile.sections.prob_classification_performance_profile_section import (
    ProbClassificationPerformanceProfileSection,
)
from evidently.model_profile.sections.regression_performance_profile_section import RegressionPerformanceProfileSection
from evidently.runner.runner import Runner
from evidently.runner.runner import RunnerOptions
from evidently.utils import NumpyEncoder


@dataclass
class ProfileRunnerOptions(RunnerOptions):
    profile_parts: Dict[str, Dict[str, str]]
    pretty_print: bool


parts_mapping = dict(
    data_drift=DataDriftProfileSection,
    cat_target_drift=CatTargetDriftProfileSection,
    classification_performance=ClassificationPerformanceProfileSection,
    prob_classification_performance=ProbClassificationPerformanceProfileSection,
    num_target_drift=NumTargetDriftProfileSection,
    regression_performance=RegressionPerformanceProfileSection,
    data_quality=DataQualityProfileSection,
)


class ProfileRunner(Runner):
    def __init__(self, options: ProfileRunnerOptions):
        super().__init__(options)
        self.options = options

    def run(self):
        (reference_data, current_data) = self._parse_data()

        parts = []

        for part, _ in self.options.profile_parts.items():
            part_class = parts_mapping.get(part, None)
            if part_class is None:
                raise ValueError(f"Unknown profile section {part}")
            parts.append(part_class())

        profile = Profile(sections=parts, options=self.options.options)
        profile.calculate(reference_data, current_data, self.options.column_mapping)
        output_path = (
            self.options.output_path
            if self.options.output_path.endswith(".json")
            else self.options.output_path + ".json"
        )

        with open(output_path, "w", encoding="utf-8") as out_file:
            json.dump(profile.object(), out_file, indent=2 if self.options.pretty_print else None, cls=NumpyEncoder)



import dataclasses
from typing import List
from typing import Optional

import numpy as np
import pandas as pd

from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.calculations.data_quality import get_rows_count
from evidently.model.widget import BaseWidgetInfo
from evidently.renderers.base_renderer import MetricRenderer
from evidently.renderers.base_renderer import default_renderer
from evidently.renderers.html_widgets import CounterData
from evidently.renderers.html_widgets import HistogramData
from evidently.renderers.html_widgets import TabData
from evidently.renderers.html_widgets import counter
from evidently.renderers.html_widgets import get_histogram_figure_with_range
from evidently.renderers.html_widgets import header_text
from evidently.renderers.html_widgets import plotly_figure
from evidently.renderers.html_widgets import table_data
from evidently.renderers.html_widgets import widget_tabs
from evidently.utils.types import Numeric
from evidently.utils.visualizations import Distribution
from evidently.utils.visualizations import get_distribution_for_column


@dataclasses.dataclass
class ValuesInRangeStat:
    number_in_range: int
    number_not_in_range: int
    share_in_range: float
    share_not_in_range: float
    # number of rows without null-like values
    number_of_values: int


@dataclasses.dataclass
class ColumnValueRangeMetricResult:
    column_name: str
    left: Numeric
    right: Numeric
    current: ValuesInRangeStat
    current_distribution: Distribution
    reference: Optional[ValuesInRangeStat] = None
    reference_distribution: Optional[Distribution] = None

    def __eq__(self, other):
        return (
            self.column_name == other.column_name
            and self.left == other.left
            and self.right == other.right
            and self.current == other.current
            and self.reference == other.reference
        )


class ColumnValueRangeMetric(Metric[ColumnValueRangeMetricResult]):
    """Calculates count and shares of values in the predefined values range"""

    column_name: str
    left: Optional[Numeric]
    right: Optional[Numeric]

    def __init__(self, column_name: str, left: Optional[Numeric] = None, right: Optional[Numeric] = None) -> None:
        self.left = left
        self.right = right
        self.column_name = column_name

    @staticmethod
    def _calculate_in_range_stats(column: pd.Series, left: Numeric, right: Numeric) -> ValuesInRangeStat:
        column = column.dropna()
        rows_count = get_rows_count(column)

        if rows_count == 0:
            number_in_range = 0
            number_not_in_range = 0
            share_in_range = 0.0
            share_not_in_range = 0.0

        else:
            number_in_range = column.between(left=float(left), right=float(right), inclusive="both").sum()
            number_not_in_range = rows_count - number_in_range
            share_in_range = number_in_range / rows_count
            share_not_in_range = number_not_in_range / rows_count

        return ValuesInRangeStat(
            number_in_range=number_in_range,
            number_not_in_range=number_not_in_range,
            share_in_range=share_in_range,
            share_not_in_range=share_not_in_range,
            number_of_values=rows_count,
        )

    def calculate(self, data: InputData) -> ColumnValueRangeMetricResult:
        if self.column_name not in data.current_data:
            raise ValueError(f"Column {self.column_name} is not in current data.")

        if not pd.api.types.is_numeric_dtype(data.current_data[self.column_name].dtype):
            raise ValueError(f"Column {self.column_name} in current data should be numeric.")

        if data.reference_data is not None:
            if self.column_name not in data.reference_data:
                raise ValueError(f"Column {self.column_name} is not in reference data.")

            if not pd.api.types.is_numeric_dtype(data.reference_data[self.column_name].dtype):
                raise ValueError(f"Column {self.column_name} in reference data should be numeric.")

        if self.left is None:
            if data.reference_data is None:
                raise ValueError("Reference should be present")

            else:
                left: Numeric = float(data.reference_data[self.column_name].min())

        else:
            left = self.left

        if self.right is None:
            if data.reference_data is None:
                raise ValueError("Reference should be present")

            else:
                right: Numeric = float(data.reference_data[self.column_name].max())

        else:
            right = self.right

        current = self._calculate_in_range_stats(data.current_data[self.column_name], left, right)

        if data.reference_data is None:
            reference_column = None
            reference = None

        else:
            reference_column = data.reference_data[self.column_name]
            reference = self._calculate_in_range_stats(data.reference_data[self.column_name], left, right)

        # calculate distribution for visualisation
        current_column = data.current_data[self.column_name]
        distributions = get_distribution_for_column(
            column_type="num",
            current=current_column,
            reference=reference_column,
        )

        return ColumnValueRangeMetricResult(
            column_name=self.column_name,
            left=left,
            right=right,
            current=current,
            reference=reference,
            current_distribution=distributions[0],
            reference_distribution=distributions[1],
        )


@default_renderer(wrap_type=ColumnValueRangeMetric)
class ColumnValueRangeMetricRenderer(MetricRenderer):
    def render_json(self, obj: ColumnValueRangeMetric) -> dict:
        result = dataclasses.asdict(obj.get_result())
        result.pop("current_distribution", None)
        result.pop("reference_distribution", None)
        return result

    @staticmethod
    def _get_table_stat(metric_result: ColumnValueRangeMetricResult) -> BaseWidgetInfo:
        matched_stat_headers = ["Metric", "Current"]
        matched_stat = [
            ["Values in range", metric_result.current.number_in_range],
            ["%", np.round(metric_result.current.share_in_range * 100, 3)],
            ["Values out of range", metric_result.current.number_not_in_range],
            ["%", np.round(metric_result.current.share_not_in_range * 100, 3)],
            ["Values count", metric_result.current.number_of_values],
        ]

        if metric_result.reference is not None:
            matched_stat_headers.append("Reference")

            matched_stat[0].append(metric_result.reference.number_in_range)
            matched_stat[1].append(np.round(metric_result.reference.share_in_range * 100, 3))
            matched_stat[2].append(metric_result.reference.number_not_in_range)
            matched_stat[3].append(np.round(metric_result.reference.share_not_in_range * 100, 3))
            matched_stat[4].append(metric_result.reference.number_of_values)

        return table_data(
            title="",
            column_names=matched_stat_headers,
            data=matched_stat,
        )

    def _get_tabs(
        self,
        metric_result: ColumnValueRangeMetricResult,
    ) -> BaseWidgetInfo:
        if metric_result.reference_distribution is not None:
            reference_histogram: Optional[HistogramData] = HistogramData(
                name="reference",
                x=list(metric_result.reference_distribution.x),
                y=list(metric_result.reference_distribution.y),
            )

        else:
            reference_histogram = None

        figure = get_histogram_figure_with_range(
            primary_hist=HistogramData(
                name="current",
                x=list(metric_result.current_distribution.x),
                y=list(metric_result.current_distribution.y),
            ),
            secondary_hist=reference_histogram,
            color_options=self.color_options,
            left=metric_result.left,
            right=metric_result.right,
        )
        figure.update_layout(
            yaxis_title="count",
            xaxis_title=metric_result.column_name,
        )

        tabs: List[TabData] = [
            TabData(
                title="Distribution",
                widget=plotly_figure(title="", figure=figure),
            ),
            TabData(
                title="Statistics",
                widget=self._get_table_stat(metric_result),
            ),
        ]
        return widget_tabs(
            title="",
            tabs=tabs,
        )

    @staticmethod
    def _get_in_range_info(stat: ValuesInRangeStat) -> str:
        percents = round(stat.share_in_range * 100, 3)
        return f"{stat.number_in_range} ({percents}%)"

    def render_html(self, obj: ColumnValueRangeMetric) -> List[BaseWidgetInfo]:
        metric_result = obj.get_result()
        column_name = metric_result.column_name

        counters = [
            CounterData.string(label="Value range", value=f"[{metric_result.left}, {metric_result.right}]"),
            CounterData.string(label="In range (current)", value=self._get_in_range_info(metric_result.current)),
        ]

        if metric_result.reference is not None:
            counters.append(
                CounterData.string(
                    label="In range (reference)", value=self._get_in_range_info(metric_result.reference)
                ),
            )

        result: List[BaseWidgetInfo] = [
            header_text(
                label=f"Column '{column_name}'. Value range.",
            ),
            counter(counters=counters),
            self._get_tabs(metric_result),
        ]
        return result



import json
from datetime import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Type

import pandas

from evidently.analyzers.base_analyzer import Analyzer
from evidently.model_profile.sections.base_profile_section import ProfileSection
from evidently.pipeline.column_mapping import ColumnMapping
from evidently.pipeline.pipeline import Pipeline
from evidently.utils import NumpyEncoder


class Profile(Pipeline):
    result: Dict[str, Any]
    stages: Sequence[ProfileSection]

    def __init__(
        self, sections: Sequence[ProfileSection], options: Optional[list] = None
    ) -> None:
        if options is None:
            options = []
        super().__init__(sections, options if options is not None else [])

    def calculate(
        self,
        reference_data: pandas.DataFrame,
        current_data: Optional[pandas.DataFrame] = None,
        column_mapping: Optional[ColumnMapping] = None,
    ) -> None:
        self.execute(reference_data, current_data, column_mapping)

    def get_analyzers(self) -> List[Type[Analyzer]]:
        return list({analyzer for tab in self.stages for analyzer in tab.analyzers()})

    def json(self) -> str:
        return json.dumps(self.object(), cls=NumpyEncoder)

    def object(self) -> Dict[str, Any]:
        result: Dict[str, Any] = {
            part.part_id(): part.get_results() for part in self.stages
        }
        result["timestamp"] = str(datetime.now())
        return result



from typing import Generator

from evidently.analyzers.num_target_drift_analyzer import NumTargetDriftAnalyzer
from evidently.calculations.data_drift import ColumnDataDriftMetrics
from evidently.model_monitoring.monitoring import MetricsType
from evidently.model_monitoring.monitoring import ModelMonitor
from evidently.model_monitoring.monitoring import ModelMonitoringMetric


class NumTargetDriftMonitorMetrics:
    """Class for numeric target grift metrics.

    Metrics list:
        - count: quantity of rows in `reference` and `current` datasets
        - drift: p_value for the data drift
        - current_correlations: correlation with `target` and `prediction` columns
            for numeric features in `current` dataset
        - reference_correlations: correlation with `target` and `prediction` columns
            for numeric features in `reference` dataset
    """

    _tag = "num_target_drift"
    count = ModelMonitoringMetric(f"{_tag}:count", ["dataset"])
    drift = ModelMonitoringMetric(f"{_tag}:drift", ["kind"])
    current_correlations = ModelMonitoringMetric(
        f"{_tag}:current_correlations", ["feature", "feature_type", "kind"]
    )
    reference_correlations = ModelMonitoringMetric(
        f"{_tag}:reference_correlations", ["feature", "feature_type", "kind"]
    )


class NumTargetDriftMonitor(ModelMonitor):
    def monitor_id(self) -> str:
        return "num_target_drift"

    def analyzers(self):
        return [NumTargetDriftAnalyzer]

    @staticmethod
    def _yield_metrics(
        metrics: ColumnDataDriftMetrics, kind: str
    ) -> Generator[MetricsType, None, None]:
        yield NumTargetDriftMonitorMetrics.drift.create(
            metrics.drift_score, dict(kind=kind)
        )

        if metrics.reference_correlations is not None:
            for (
                feature_name,
                correlation_value,
            ) in metrics.reference_correlations.items():
                yield NumTargetDriftMonitorMetrics.reference_correlations.create(
                    correlation_value,
                    dict(feature=feature_name, feature_type="num", kind=kind),
                )

        if metrics.current_correlations is not None:
            for feature_name, correlation_value in metrics.current_correlations.items():
                yield NumTargetDriftMonitorMetrics.current_correlations.create(
                    correlation_value,
                    dict(feature=feature_name, feature_type="num", kind=kind),
                )

    def metrics(self, analyzer_results):
        results = NumTargetDriftAnalyzer.get_results(analyzer_results)

        # quantity of rows in income data
        yield NumTargetDriftMonitorMetrics.count.create(
            results.reference_data_count, dict(dataset="reference")
        )
        yield NumTargetDriftMonitorMetrics.count.create(
            results.current_data_count, dict(dataset="current")
        )

        if results.prediction_metrics:
            for metric in self._yield_metrics(
                metrics=results.prediction_metrics, kind="prediction"
            ):
                yield metric

        if results.target_metrics:
            for metric in self._yield_metrics(
                metrics=results.target_metrics, kind="target"
            ):
                yield metric



"""Epps-Singleton test of two samples.

Name: "es"

Import:

    >>> from evidently.calculations.stattests import epps_singleton_test

Properties:
- only for numerical features
- returns p-value
- default threshold 0.05

Example:
    Using by object:

    >>> from evidently.options import DataDriftOptions
    >>> from evidently.calculations.stattests import epps_singleton_test
    >>> options = DataDriftOptions(all_features_stattest=epps_singleton_test)

    Using by name:

    >>> from evidently.options import DataDriftOptions
    >>> options = DataDriftOptions(all_features_stattest="es")
"""
from typing import Tuple

import numpy as np
import pandas as pd
from scipy.stats import epps_singleton_2samp
from scipy.stats import iqr

from evidently.calculations.stattests.registry import StatTest
from evidently.calculations.stattests.registry import register_stattest


def _epps_singleton(
    reference_data: pd.Series,
    current_data: pd.Series,
    feature_type: str,
    threshold: float,
) -> Tuple[float, bool]:
    """Run the Epps-Singleton (ES) test of two samples.
    Args:
        reference_data: reference data
        current_data: current data
        threshold: level of significance (default will be 0.05)
    Returns:
        p_value: p-value based on the asymptotic chi2-distribution.
        test_result: whether the drift is detected
    """
    # raised error if iqr is zero
    iqr_value = iqr(np.hstack((reference_data, current_data)))
    if iqr_value == 0:
        raise ValueError(
            f'{"Epps-Singleton testsinterquartile range (IQR) is zero. IQR of 0 means there is no variability in the middle 50% of your data, but the center of the distribution can be anywhere. Expected: your data sould have higher variability/variance."}'
        )
    p_value = epps_singleton_2samp(reference_data, current_data)[1]
    return p_value, p_value < threshold


epps_singleton_test = StatTest(
    name="es",
    display_name="Epps-Singleton",
    func=_epps_singleton,
    allowed_feature_types=["num"],
    default_threshold=0.05,
)

register_stattest(epps_singleton_test)



#!/usr/bin/env python
# coding: utf-8
import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from evidently import ColumnMapping
from evidently.analyzers.data_quality_analyzer import DataQualityAnalyzer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo


class DataQualityCorrelationsWidget(Widget):
    period_prefix: str

    def analyzers(self):
        return [DataQualityAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        data_quality_results = DataQualityAnalyzer.get_results(analyzers_results)
        is_current_data = current_data is not None
        reference_correlations = data_quality_results.reference_correlations
        if is_current_data:
            current_correlations = data_quality_results.current_correlations
        else:
            current_correlations = None

        additional_graphs = []
        parts = []
        for kind in ["pearson", "spearman", "kendall", "cramer_v"]:
            if reference_correlations[kind].shape[0] > 1:
                correlation_figure = self._plot_correlation_figure(kind, reference_correlations, current_correlations)
                additional_graphs.append(
                    AdditionalGraphInfo(
                        kind,
                        {
                            "data": correlation_figure["data"],
                            "layout": correlation_figure["layout"],
                        },
                    )
                )

                parts.append({"title": kind, "id": kind})

        if is_current_data:
            metrics_values_headers = [
                "top 5 correlation diff category (Cramer_V)",
                "value ref",
                "value curr",
                "difference",
                "top 5 correlation diff numerical (Spearman)",
                "value ref",
                "value curr",
                "difference",
            ]

        else:
            metrics_values_headers = [
                "top 5 correlated category (Cramer_V)",
                "Value",
                "top 5 correlated numerical (Spearman)",
                "Value",
            ]

        metrics = self._make_metrics(reference_correlations, current_correlations)

        wi = BaseWidgetInfo(
            type="rich_data",
            title="",
            size=2,
            params={
                "header": "Correlations",
                "description": "",
                "metricsValuesHeaders": metrics_values_headers,
                "metrics": metrics,
                "details": {"parts": parts, "insights": []},
            },
            additionalGraphs=additional_graphs,
        )

        return wi

    def _plot_correlation_figure(
        self, kind: str, reference_correlations: dict, current_correlations: Optional[dict]
    ) -> dict:
        columns = reference_correlations[kind].columns
        heatmap_text = None
        heatmap_texttemplate = None
        if current_correlations is not None:
            cols = 2
            subplot_titles = ["reference", "current"]
        else:
            cols = 1
            subplot_titles = [""]

        fig = make_subplots(rows=1, cols=cols, subplot_titles=subplot_titles, shared_yaxes=True)
        if len(columns) < 15:
            heatmap_text = np.round(reference_correlations[kind], 2).astype(str)
            heatmap_texttemplate = "%{text}"

        trace = go.Heatmap(
            z=reference_correlations[kind],
            x=columns,
            y=columns,
            text=heatmap_text,
            texttemplate=heatmap_texttemplate,
            coloraxis="coloraxis",
        )
        fig.add_trace(trace, 1, 1)
        if current_correlations is not None:
            if len(columns) < 15:
                heatmap_text = np.round(current_correlations[kind], 2).astype(str)
                heatmap_texttemplate = "%{text}"

            trace = go.Heatmap(
                z=current_correlations[kind],
                x=columns,
                y=columns,
                text=heatmap_text,
                texttemplate=heatmap_texttemplate,
                coloraxis="coloraxis",
            )
            fig.add_trace(trace, 1, 2)
        fig.update_layout(coloraxis={"colorscale": "RdBu_r"})
        correlation_figure = json.loads(fig.to_json())
        return correlation_figure

    def _get_df_corr_features_sorted(self, df_corr: pd.DataFrame) -> pd.DataFrame:
        df_corr = df_corr.stack().reset_index()
        df_corr.columns = ["col_1", "col_2", "value"]
        df_corr["value"] = np.round(df_corr["value"], 3)
        df_corr["abs_value"] = np.abs(df_corr["value"])
        df_corr = df_corr.sort_values(["col_1", "col_2"])
        df_corr["keep"] = df_corr["col_1"].str.get(0) < df_corr["col_2"].str.get(0)
        df_corr = df_corr[df_corr["keep"]]
        df_corr = df_corr.sort_values("abs_value", ascending=False)
        df_corr["features"] = df_corr["col_1"] + ", " + df_corr["col_2"]
        return df_corr[["features", "value"]]

    def _get_rel_diff_corr_features_sorted(self, ref_corr: pd.DataFrame, curr_corr: pd.DataFrame) -> pd.DataFrame:
        ref_corr = self._get_df_corr_features_sorted(ref_corr).rename(columns={"value": "value_ref"})
        curr_corr = self._get_df_corr_features_sorted(curr_corr).rename(columns={"value": "value_curr"})
        com_corr = ref_corr.merge(curr_corr, on="features", how="left")
        com_corr["value_diff"] = np.round((com_corr["value_ref"] - com_corr["value_curr"]), 3)
        com_corr["abs_value_diff"] = np.abs(com_corr["value_diff"])
        com_corr = com_corr.sort_values("abs_value_diff", ascending=False)
        return com_corr[["features", "value_ref", "value_curr", "value_diff"]]

    def _make_metrics(self, reference_correlations: dict, current_correlations: Optional[dict]):
        metrics = []
        if current_correlations is not None:
            if reference_correlations["spearman"].shape[0] > 1:
                com_num_corr = self._get_rel_diff_corr_features_sorted(
                    reference_correlations["spearman"], current_correlations["spearman"]
                )
            else:
                com_num_corr = pd.DataFrame()
            if reference_correlations["cramer_v"].shape[0] > 1:
                com_cat_corr = self._get_rel_diff_corr_features_sorted(
                    reference_correlations["cramer_v"], current_correlations["cramer_v"]
                )
            else:
                com_cat_corr = pd.DataFrame()
            for i in range(5):
                values = ["-", "-", "-", "-", "-", "-", "-", "-"]
                if i < com_cat_corr.shape[0]:
                    values[0] = com_cat_corr.iloc[i, 0]
                    values[1] = com_cat_corr.iloc[i, 1]
                    values[2] = com_cat_corr.iloc[i, 2]
                    values[3] = com_cat_corr.iloc[i, 3]
                if i < com_num_corr.shape[0]:
                    values[4] = com_num_corr.iloc[i, 0]
                    values[5] = com_num_corr.iloc[i, 1]
                    values[6] = com_num_corr.iloc[i, 2]
                    values[7] = com_num_corr.iloc[i, 3]
                metrics.append(
                    {"label": "", "values": values},
                )
        else:
            if reference_correlations["spearman"].shape[0] > 0:
                ref_num_corr = self._get_df_corr_features_sorted(reference_correlations["spearman"])
            else:
                ref_num_corr = pd.DataFrame()
            if reference_correlations["cramer_v"].shape[0] > 0:
                ref_cat_corr = self._get_df_corr_features_sorted(reference_correlations["cramer_v"])
            else:
                ref_cat_corr = pd.DataFrame()
            for i in range(5):
                values = ["-", "-", "-", "-"]
                if i < ref_cat_corr.shape[0]:
                    values[0] = ref_cat_corr.iloc[i, 0]
                    values[1] = ref_cat_corr.iloc[i, 1]
                if i < ref_num_corr.shape[0]:
                    values[2] = ref_num_corr.iloc[i, 0]
                    values[3] = ref_num_corr.iloc[i, 1]
                metrics.append(
                    {"label": "", "values": values},
                )
        return metrics



import pandas as pd


def make_target_bins_for_reg_plots(
    curr: pd.DataFrame, target_column, preds_column, ref: pd.DataFrame = None
) -> pd.DataFrame:
    df_for_bins = pd.DataFrame(
        {
            "data": "curr",
            target_column: curr[target_column],
            preds_column: curr[preds_column],
        }
    )
    if ref is not None:
        df_for_bins = pd.concat(
            [
                df_for_bins,
                pd.DataFrame(
                    {
                        "data": "ref",
                        target_column: ref[target_column],
                        preds_column: ref[preds_column],
                    }
                ),
            ]
        )
    df_for_bins["target_binned"] = pd.cut(
        df_for_bins[target_column], min(df_for_bins[target_column].nunique(), 10)
    )
    return df_for_bins


def apply_func_to_binned_data(
    df_for_bins, func, target_column, preds_column, is_ref_data=False
):
    result = {}

    def _apply(x):
        if x.shape[0] == 0:
            return None
        return func(x[target_column], x[preds_column])

    result["current"] = (
        df_for_bins[df_for_bins.data == "curr"].groupby("target_binned").apply(_apply)
    )

    if is_ref_data:
        result["reference"] = (
            df_for_bins[df_for_bins.data == "ref"]
            .groupby("target_binned")
            .apply(_apply)
        )
    return result



from evidently.utils.dashboard import file_html_template
from evidently.utils.dashboard import inline_template


def determine_template(mode: str):
    # pylint: disable=import-outside-toplevel
    render_mode = mode
    try:
        from IPython import get_ipython

        is_colab = type(get_ipython()).__module__.startswith("google.colab")
    except ImportError:
        is_colab = False

    if mode == "auto":
        if is_colab:
            render_mode = "inline"
        else:
            render_mode = "nbextension"
    if render_mode == "inline":
        return file_html_template
    if render_mode == "nbextension":
        return inline_template
    raise ValueError(f"Unexpected value {mode}/{render_mode} for mode")



from typing import Dict
from typing import List
from typing import Optional

from evidently import TaskType
from evidently.base_metric import InputData
from evidently.calculations.stattests import PossibleStatTestType
from evidently.test_preset.test_preset import TestPreset
from evidently.tests import TestAllColumnsShareOfMissingValues
from evidently.tests import TestCatColumnsOutOfListValues
from evidently.tests import TestColumnDrift
from evidently.tests import TestColumnsType
from evidently.tests import TestNumColumnsMeanInNSigmas
from evidently.tests import TestNumColumnsOutOfRangeValues
from evidently.tests import TestShareOfDriftedColumns
from evidently.utils.data_drift_utils import resolve_stattest_threshold
from evidently.utils.data_operations import DatasetColumns


class NoTargetPerformanceTestPreset(TestPreset):
    """
    No Target Performance tests.

    Args:
        columns: list of columns include to tests

    Contains tests:
    - `TestColumnValueDrift`
    - `TestShareOfDriftedColumns`
    - `TestColumnsType`
    - `TestAllColumnsShareOfMissingValues`
    - `TestNumColumnsOutOfRangeValues`
    - `TestCatColumnsOutOfListValues`
    - `TestNumColumnsMeanInNSigmas`
    - `TestCustomFeaturesValueDrift`
    """

    columns: Optional[List[str]]
    drift_share: Optional[float]
    stattest: Optional[PossibleStatTestType] = None
    cat_stattest: Optional[PossibleStatTestType] = None
    num_stattest: Optional[PossibleStatTestType] = None
    text_stattest: Optional[PossibleStatTestType] = None
    per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None
    stattest_threshold: Optional[float] = None
    cat_stattest_threshold: Optional[float] = None
    num_stattest_threshold: Optional[float] = None
    text_stattest_threshold: Optional[float] = None
    per_column_stattest_threshold: Optional[Dict[str, float]] = None

    def __init__(
        self,
        columns: Optional[List[str]] = None,
        drift_share: Optional[float] = None,
        stattest: Optional[PossibleStatTestType] = None,
        cat_stattest: Optional[PossibleStatTestType] = None,
        num_stattest: Optional[PossibleStatTestType] = None,
        text_stattest: Optional[PossibleStatTestType] = None,
        per_column_stattest: Optional[Dict[str, PossibleStatTestType]] = None,
        stattest_threshold: Optional[float] = None,
        cat_stattest_threshold: Optional[float] = None,
        num_stattest_threshold: Optional[float] = None,
        text_stattest_threshold: Optional[float] = None,
        per_column_stattest_threshold: Optional[Dict[str, float]] = None,
    ):
        super().__init__()
        self.columns = columns
        self.drift_share = drift_share
        self.stattest = stattest
        self.cat_stattest = cat_stattest
        self.num_stattest = num_stattest
        self.text_stattest = text_stattest
        self.per_column_stattest = per_column_stattest
        self.stattest_threshold = stattest_threshold
        self.cat_stattest_threshold = cat_stattest_threshold
        self.num_stattest_threshold = num_stattest_threshold
        self.text_stattest_threshold = text_stattest_threshold
        self.per_feature_threshold = per_column_stattest_threshold

    def generate_tests(self, data: InputData, columns: DatasetColumns):
        preset_tests: List = []

        if columns.utility_columns.prediction is not None and isinstance(columns.utility_columns.prediction, str):
            stattest, threshold = resolve_stattest_threshold(
                columns.utility_columns.prediction,
                "cat" if columns.task == TaskType.CLASSIFICATION_TASK else "num",
                self.stattest,
                self.cat_stattest,
                self.num_stattest,
                self.text_stattest,
                self.per_column_stattest,
                self.stattest_threshold,
                self.cat_stattest_threshold,
                self.num_stattest_threshold,
                self.text_stattest_threshold,
                self.per_column_stattest_threshold,
            )
            preset_tests.append(
                TestColumnDrift(
                    column_name=columns.utility_columns.prediction,
                    stattest=stattest,
                    stattest_threshold=threshold,
                )
            )
        preset_tests.append(
            TestShareOfDriftedColumns(
                lt=0.3 if self.drift_share is None else self.drift_share,
                stattest=self.stattest,
                cat_stattest=self.cat_stattest,
                num_stattest=self.num_stattest,
                text_stattest=self.text_stattest,
                per_column_stattest=self.per_column_stattest,
                stattest_threshold=self.stattest_threshold,
                cat_stattest_threshold=self.cat_stattest_threshold,
                num_stattest_threshold=self.num_stattest_threshold,
                text_stattest_threshold=self.text_stattest_threshold,
                per_column_stattest_threshold=self.per_feature_threshold,
            )
        )
        preset_tests.append(TestColumnsType())
        preset_tests.append(TestAllColumnsShareOfMissingValues(columns=self.columns))
        preset_tests.append(TestNumColumnsOutOfRangeValues(columns=self.columns))
        preset_tests.append(TestCatColumnsOutOfListValues(columns=self.columns))
        preset_tests.append(TestNumColumnsMeanInNSigmas(columns=self.columns))

        return preset_tests



import warnings

from .monitoring import ModelMonitoring
from .monitors.cat_target_drift import CatTargetDriftMonitor
from .monitors.classification_performance import ClassificationPerformanceMonitor
from .monitors.data_drift import DataDriftMonitor
from .monitors.data_quality import DataQualityMonitor
from .monitors.num_target_drift import NumTargetDriftMonitor
from .monitors.prob_classification_performance import ProbClassificationPerformanceMonitor
from .monitors.regression_performance import RegressionPerformanceMonitor

warnings.warn("model monitoring is deprecated, use metrics instead")



from datetime import datetime
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Optional
from typing import Type

from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.cat_target_drift_analyzer import CatTargetDriftAnalyzer
from evidently.model_profile.sections.base_profile_section import ProfileSection


class CatTargetDriftProfileSection(ProfileSection):
    def part_id(self) -> str:
        return "cat_target_drift"

    def __init__(self) -> None:
        super().__init__()
        self.analyzers_types = [CatTargetDriftAnalyzer]
        self._result = None

    def analyzers(self) -> Iterable[Type[Analyzer]]:
        return self.analyzers_types

    def calculate(
        self, reference_data, current_data, column_mapping, analyzers_results
    ) -> None:
        result = CatTargetDriftAnalyzer.get_results(analyzers_results)
        result_json: Dict[str, Any] = result.columns.as_dict()
        result_json["metrics"] = {}

        if result.target_metrics:
            result_json["metrics"]["target_name"] = result.target_metrics.column_name
            result_json["metrics"]["target_type"] = "cat"
            result_json["metrics"]["target_drift"] = result.target_metrics.drift_score

        if result.prediction_metrics:
            result_json["metrics"][
                "prediction_name"
            ] = result.prediction_metrics.column_name
            result_json["metrics"]["prediction_type"] = "cat"
            result_json["metrics"][
                "prediction_drift"
            ] = result.prediction_metrics.drift_score

        self._result = {
            "name": self.part_id(),
            "datetime": str(datetime.now()),
            "data": result_json,
        }

    def get_results(self) -> Optional[dict]:
        return self._result



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.figure_factory as ff

from evidently import ColumnMapping
from evidently.analyzers.prob_distribution_analyzer import ProbDistributionAnalyzer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class ProbClassPredDistrWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [ProbDistributionAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = ProbDistributionAnalyzer.get_results(analyzers_results)
        utility_columns = results.columns.utility_columns

        if (
            isinstance(utility_columns.prediction, str)
            or utility_columns.prediction is None
        ):
            return None

        if utility_columns.target is None or utility_columns.prediction is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )

            return None

        if self.dataset == "current":
            dataset_to_plot = (
                current_data.copy(deep=False) if current_data is not None else None
            )

        else:
            dataset_to_plot = reference_data.copy(deep=False)

        if dataset_to_plot is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires reference dataset but it is None"
                )

            return None

        dataset_to_plot.replace([np.inf, -np.inf], np.nan, inplace=True)

        # plot distributions
        graphs = []

        for label in utility_columns.prediction:
            pred_distr = ff.create_distplot(
                [
                    dataset_to_plot[dataset_to_plot[utility_columns.target] == label][
                        label
                    ],
                    dataset_to_plot[dataset_to_plot[utility_columns.target] != label][
                        label
                    ],
                ],
                [str(label), "other"],
                colors=[color_options.primary_color, color_options.secondary_color],
                bin_size=0.05,
                show_curve=False,
                show_rug=True,
            )

            pred_distr.update_layout(
                xaxis_title="Probability",
                yaxis_title="Share",
                legend=dict(
                    orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
                ),
            )

            pred_distr_json = json.loads(pred_distr.to_json())

            graphs.append(
                {
                    "id": "tab_" + str(label),
                    "title": str(label),
                    "graph": {
                        "data": pred_distr_json["data"],
                        "layout": pred_distr_json["layout"],
                    },
                }
            )

        return BaseWidgetInfo(
            title=self.title,
            type="tabbed_graph",
            size=1 if current_data is not None else 2,
            params={"graphs": graphs},
        )



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import pandas as pd
import plotly.express as px

from evidently import ColumnMapping
from evidently.analyzers.cat_target_drift_analyzer import CatTargetDriftAnalyzer
from evidently.dashboard.widgets.utils import CutQuantileTransformer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import AdditionalGraphInfo
from evidently.model.widget import BaseWidgetInfo
from evidently.options import QualityMetricsOptions


class CatTargetPredFeatureTable(Widget):
    @staticmethod
    def _get_rows_per_page(columns) -> int:
        return min(len(columns.num_feature_names) + len(columns.cat_feature_names), 10)

    def analyzers(self):
        return [CatTargetDriftAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:

        results = CatTargetDriftAnalyzer.get_results(analyzers_results)
        quality_metrics_options = self.options_provider.get(QualityMetricsOptions)
        cut_quantile = quality_metrics_options.cut_quantile

        if current_data is None:
            raise ValueError("current_data should be present")

        target_name = results.columns.utility_columns.target
        prediction_name = None
        if results.prediction_metrics:
            prediction_name = results.prediction_metrics.column_name

        if prediction_name is not None and target_name is not None:
            additional_graphs_data = []
            params_data = []
            for feature_name in results.columns.get_all_features_list(
                cat_before_num=False
            ):
                # add data for table in params
                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {
                                    "title": "Target",
                                    "id": feature_name + "_target_values",
                                },
                                {
                                    "title": "Prediction",
                                    "id": feature_name + "_prediction_values",
                                },
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                    }
                )

                # create target plot
                reference_data["dataset"] = "Reference"
                current_data["dataset"] = "Current"
                quantile = quality_metrics_options.get_cut_quantile(feature_name)
                if cut_quantile and quantile is not None:
                    side, q = quantile
                    cqt = CutQuantileTransformer(side=side, q=q)
                    cqt.fit(reference_data[feature_name])
                    reference_data_to_plot = cqt.transform_df(
                        reference_data, feature_name
                    )
                    current_data_to_plot = cqt.transform_df(current_data, feature_name)
                else:
                    reference_data_to_plot = reference_data
                    current_data_to_plot = current_data
                merged_data = pd.concat([reference_data_to_plot, current_data_to_plot])

                target_fig = px.histogram(
                    merged_data,
                    x=feature_name,
                    color=target_name,
                    facet_col="dataset",
                    barmode="overlay",
                    category_orders={"dataset": ["Reference", "Current"]},
                )

                target_fig_json = json.loads(target_fig.to_json())

                # create prediction plot
                pred_fig = px.histogram(
                    merged_data,
                    x=feature_name,
                    color=prediction_name,
                    facet_col="dataset",
                    barmode="overlay",
                    category_orders={"dataset": ["Reference", "Current"]},
                )

                pred_fig_json = json.loads(pred_fig.to_json())

                # write plot data in table as additional data
                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_target_values",
                        {
                            "data": target_fig_json["data"],
                            "layout": target_fig_json["layout"],
                        },
                    )
                )

                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_prediction_values",
                        {
                            "data": pred_fig_json["data"],
                            "layout": pred_fig_json["layout"],
                        },
                    )
                )

            return BaseWidgetInfo(
                title=self.title,
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": self._get_rows_per_page(results.columns),
                    "columns": [{"title": "Feature", "field": "f1"}],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )

        if target_name is not None:
            additional_graphs_data = []
            params_data = []

            for feature_name in results.columns.get_all_features_list(
                cat_before_num=False
            ):
                # add data for table in params
                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {
                                    "title": "Target",
                                    "id": feature_name + "_target_values",
                                }
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                    }
                )

                # create target plot
                # TO DO%: out pf the cycle
                reference_data["dataset"] = "Reference"
                current_data["dataset"] = "Current"
                quantile = quality_metrics_options.get_cut_quantile(feature_name)
                if cut_quantile and quantile is not None:
                    side, q = quantile
                    cqt = CutQuantileTransformer(side=side, q=q)
                    cqt.fit(reference_data[feature_name])
                    reference_data_to_plot = cqt.transform_df(
                        reference_data, feature_name
                    )
                    current_data_to_plot = cqt.transform_df(current_data, feature_name)
                else:
                    reference_data_to_plot = reference_data
                    current_data_to_plot = current_data
                merged_data = pd.concat([reference_data_to_plot, current_data_to_plot])

                target_fig = px.histogram(
                    merged_data,
                    x=feature_name,
                    color=target_name,
                    facet_col="dataset",
                    barmode="overlay",
                    category_orders={"dataset": ["Reference", "Current"]},
                )

                target_fig_json = json.loads(target_fig.to_json())

                # write plot data in table as additional data
                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_target_values",
                        {
                            "data": target_fig_json["data"],
                            "layout": target_fig_json["layout"],
                        },
                    )
                )

            return BaseWidgetInfo(
                title=self.title,
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": self._get_rows_per_page(results.columns),
                    "columns": [{"title": "Feature", "field": "f1"}],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )
        if results.columns.utility_columns.prediction is not None:
            additional_graphs_data = []
            params_data = []
            for feature_name in (
                results.columns.num_feature_names + results.columns.cat_feature_names
            ):
                # add data for table in params
                params_data.append(
                    {
                        "details": {
                            "parts": [
                                {
                                    "title": "Prediction",
                                    "id": feature_name + "_prediction_values",
                                }
                            ],
                            "insights": [],
                        },
                        "f1": feature_name,
                    }
                )

                # create target plot
                reference_data["dataset"] = "Reference"
                current_data["dataset"] = "Current"
                quantile = quality_metrics_options.get_cut_quantile(feature_name)
                if cut_quantile and quantile is not None:
                    side, q = quantile
                    cqt = CutQuantileTransformer(side=side, q=q)
                    cqt.fit(reference_data[feature_name])
                    reference_data_to_plot = cqt.transform_df(
                        reference_data, feature_name
                    )
                    current_data_to_plot = cqt.transform_df(current_data, feature_name)
                else:
                    reference_data_to_plot = reference_data
                    current_data_to_plot = current_data
                merged_data = pd.concat([reference_data_to_plot, current_data_to_plot])

                prediction_fig = px.histogram(
                    merged_data,
                    x=feature_name,
                    barmode="overlay",
                    color=results.columns.utility_columns.prediction,
                    facet_col="dataset",
                    category_orders={"dataset": ["Reference", "Current"]},
                )

                prediction_fig_json = json.loads(prediction_fig.to_json())

                # write plot data in table as additional data
                additional_graphs_data.append(
                    AdditionalGraphInfo(
                        feature_name + "_prediction_values",
                        {
                            "data": prediction_fig_json["data"],
                            "layout": prediction_fig_json["layout"],
                        },
                    )
                )

            return BaseWidgetInfo(
                title=self.title,
                type="big_table",
                size=2,
                params={
                    "rowsPerPage": self._get_rows_per_page(results.columns),
                    "columns": [{"title": "Feature", "field": "f1"}],
                    "data": params_data,
                },
                additionalGraphs=additional_graphs_data,
            )
        raise ValueError(
            f"Widget {self.title} require 'prediction' or 'target' columns not to be None"
        )



#!/usr/bin/env python
# coding: utf-8

from typing import List
from typing import Optional

import numpy as np
import pandas as pd

from evidently import ColumnMapping
from evidently.analyzers.data_quality_analyzer import DataQualityAnalyzer
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo


class DataQualitySummaryWidget(Widget):
    def analyzers(self):
        return [DataQualityAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        data_quality_results = DataQualityAnalyzer.get_results(analyzers_results)
        reference_stats = self._get_df_stats(
            data_quality_results,
            reference_data,
            data_quality_results.reference_features_stats,
        )
        if current_data is not None:
            current_stats = self._get_df_stats(
                data_quality_results,
                current_data,
                data_quality_results.current_features_stats,
            )
        else:
            current_stats = None

        if current_data is not None:
            metrics_values_headers = ["reference", "current"]

        else:
            metrics_values_headers = [""]

        stats_list = [
            "target column",
            "date column",
            "number of variables",
            "number of observations",
            "missing cells",
            "categorical features",
            "numeric features",
            "datetime features",
            "constant features",
            "empty features",
            "almost constant features",
            "almost empty features",
        ]
        metrics = self._get_stats_with_names(stats_list, reference_stats, current_stats)

        wi = BaseWidgetInfo(
            type="rich_data",
            title="",
            size=2,
            params={
                "header": "Data Summary",
                "description": "",
                "metricsValuesHeaders": metrics_values_headers,
                "metrics": metrics,
            },
        )

        return wi

    def _get_df_stats(self, data_quality_results, df, df_stats):
        result = {}
        all_features = data_quality_results.columns.get_all_features_list(
            cat_before_num=True, include_datetime_feature=True
        )
        if data_quality_results.columns.utility_columns.target:
            target_name = data_quality_results.columns.utility_columns.target
            # target_type = df_stats[target_name].feature_type
            # all_features = [target_name] + all_features
        else:
            target_name = None
        if data_quality_results.columns.utility_columns.date:
            date_name = data_quality_results.columns.utility_columns.date
            all_features = [date_name] + all_features
        else:
            date_name = None

        if target_name:
            result["target column"] = target_name
        else:
            result["target column"] = "None"
        if date_name:
            result["date column"] = date_name
        else:
            result["date column"] = "None"
        result["number of variables"] = len(all_features)
        result["number of observations"] = df.shape[0]
        missing_cells = df[all_features].isnull().sum().sum()
        missing_cells_percentage = np.round(
            missing_cells
            / (result["number of variables"] * result["number of observations"]),
            2,
        )
        result["missing cells"] = f"{missing_cells} ({missing_cells_percentage}%)"
        result["categorical features"] = len(
            data_quality_results.columns.cat_feature_names
        )
        result["numeric features"] = len(data_quality_results.columns.num_feature_names)
        result["datetime features"] = len(
            data_quality_results.columns.datetime_feature_names
        )
        if date_name:
            result["datetime features"] = result["datetime features"] + 1
        constant_values = pd.Series(
            [df_stats[x].most_common_value_percentage for x in all_features]
        )
        empty_values = pd.Series([df_stats[x].missing_percentage for x in all_features])
        result["constant features"] = (constant_values == 100).sum()
        result["empty features"] = (empty_values == 100).sum()
        result["almost constant features"] = (constant_values >= 95).sum()
        result["almost empty features"] = (empty_values >= 95).sum()
        return result

    @staticmethod
    def _get_stats_with_names(
        stats_list: List[str],
        reference_stats: dict,
        current_stats: Optional[dict],
    ) -> List[dict]:

        result = []

        for val in stats_list:
            values = [reference_stats[val]]

            if current_stats is not None:
                values.append(current_stats[val])

            result.append(
                {
                    "label": val,
                    "values": values,
                }
            )
        return result



from evidently.base_metric import InputData
from evidently.metric_preset.metric_preset import MetricPreset
from evidently.metrics import ColumnDriftMetric
from evidently.metrics import ColumnSummaryMetric
from evidently.metrics import TextDescriptorsCorrelationMetric
from evidently.metrics import TextDescriptorsDistribution
from evidently.metrics import TextDescriptorsDriftMetric
from evidently.utils.data_operations import DatasetColumns


class TextOverviewPreset(MetricPreset):
    """Metric preset for text column analysis.

    Contains metrics:
    - ColumnSummaryMetric
    - TextDescriptorsDistribution
    - TextDescriptorsCorrelation
    - ColumnDriftMetric
    - TextDescriptorsDescriptorsDriftMetric

    Args:
        column_name: text column name.
    """

    column_name: str

    def __init__(self, column_name: str):
        super().__init__()
        self.column_name = column_name

    def generate_metrics(self, data: InputData, columns: DatasetColumns):
        result = [
            ColumnSummaryMetric(column_name=self.column_name),
            TextDescriptorsDistribution(column_name=self.column_name),
            TextDescriptorsCorrelationMetric(column_name=self.column_name),
        ]
        if data.reference_data is not None:
            result.extend(
                [
                    ColumnDriftMetric(column_name=self.column_name),
                    TextDescriptorsDriftMetric(column_name=self.column_name),
                ]
            )
        return result



#!/usr/bin/env python
# coding: utf-8
from dataclasses import dataclass
from typing import Dict
from typing import List
from typing import Optional
from typing import Union

import numpy as np
import pandas as pd
from sklearn import metrics

from evidently import ColumnMapping
from evidently.analyzers.base_analyzer import Analyzer
from evidently.analyzers.base_analyzer import BaseAnalyzerResult
from evidently.calculations.classification_performance import calculate_confusion_by_classes
from evidently.options import QualityMetricsOptions
from evidently.utils.data_operations import process_columns


@dataclass
class ConfusionMatrix:
    labels: List[str]
    values: list


@dataclass
class ProbClassificationPerformanceMetrics:
    """Class for performance metrics values"""

    accuracy: float
    precision: float
    recall: float
    f1: float
    roc_auc: float
    log_loss: float
    metrics_matrix: dict
    confusion_matrix: ConfusionMatrix
    confusion_by_classes: Dict[Union[str, int], Dict[str, int]]
    roc_aucs: Optional[list] = None
    roc_curve: Optional[dict] = None
    pr_curve: Optional[dict] = None
    pr_table: Optional[Union[dict, list]] = None


@dataclass
class ProbClassificationPerformanceAnalyzerResults(BaseAnalyzerResult):
    quality_metrics_options: QualityMetricsOptions
    reference_metrics: Optional[ProbClassificationPerformanceMetrics] = None
    current_metrics: Optional[ProbClassificationPerformanceMetrics] = None


class ProbClassificationPerformanceAnalyzer(Analyzer):
    @staticmethod
    def get_results(analyzer_results) -> ProbClassificationPerformanceAnalyzerResults:
        return analyzer_results[ProbClassificationPerformanceAnalyzer]

    def calculate(
        self, reference_data: pd.DataFrame, current_data: Optional[pd.DataFrame], column_mapping: ColumnMapping
    ) -> ProbClassificationPerformanceAnalyzerResults:
        if reference_data is None:
            raise ValueError("reference_data should be present")

        columns = process_columns(reference_data, column_mapping)
        target_column = columns.utility_columns.target
        prediction_column = columns.utility_columns.prediction
        quality_metrics_options = self.options_provider.get(QualityMetricsOptions)
        result = ProbClassificationPerformanceAnalyzerResults(
            columns=columns,
            quality_metrics_options=quality_metrics_options,
        )
        classification_threshold = quality_metrics_options.classification_threshold

        if target_column is not None and prediction_column is not None:
            target_and_preds = [target_column]
            if isinstance(prediction_column, str):
                target_and_preds += [prediction_column]
            else:
                target_and_preds += prediction_column
            reference_data.replace([np.inf, -np.inf], np.nan, inplace=True)
            reference_data.dropna(axis=0, how="any", inplace=True, subset=target_and_preds)
            binaraized_target = (reference_data[target_column].values.reshape(-1, 1) == prediction_column).astype(int)
            array_prediction = reference_data[prediction_column].to_numpy()

            if len(prediction_column) > 2:
                prediction_ids = np.argmax(array_prediction, axis=-1)
                prediction_labels = [prediction_column[x] for x in prediction_ids]

            else:
                maper = {True: prediction_column[0], False: prediction_column[1]}
                prediction_labels = (reference_data[prediction_column[0]] >= classification_threshold).map(maper)

            labels = sorted(set(reference_data[target_column]))

            # calculate quality metrics
            roc_auc = metrics.roc_auc_score(binaraized_target, array_prediction, average="macro")
            log_loss = metrics.log_loss(binaraized_target, array_prediction)
            accuracy_score = metrics.accuracy_score(reference_data[target_column], prediction_labels)
            avg_precision = metrics.precision_score(reference_data[target_column], prediction_labels, average="macro")
            avg_recall = metrics.recall_score(reference_data[target_column], prediction_labels, average="macro")
            avg_f1 = metrics.f1_score(reference_data[target_column], prediction_labels, average="macro")

            # calculate class support and metrics matrix
            metrics_matrix = metrics.classification_report(
                reference_data[target_column], prediction_labels, output_dict=True
            )

            roc_aucs = None

            if len(prediction_column) > 2:
                roc_aucs = metrics.roc_auc_score(binaraized_target, array_prediction, average=None).tolist()

            # calculate confusion matrix
            conf_matrix = metrics.confusion_matrix(reference_data[target_column], prediction_labels)
            confusion_by_classes = calculate_confusion_by_classes(conf_matrix, labels)

            result.reference_metrics = ProbClassificationPerformanceMetrics(
                accuracy=accuracy_score,
                precision=avg_precision,
                recall=avg_recall,
                f1=avg_f1,
                roc_auc=roc_auc,
                log_loss=log_loss,
                metrics_matrix=metrics_matrix,
                confusion_matrix=ConfusionMatrix(labels=labels, values=conf_matrix.tolist()),
                roc_aucs=roc_aucs,
                confusion_by_classes=confusion_by_classes,
            )

            # calculate ROC and PR curves, PR table
            if len(prediction_column) <= 2:
                binaraized_target = pd.DataFrame(binaraized_target[:, 0])
                binaraized_target.columns = ["target"]

                fpr, tpr, thrs = metrics.roc_curve(binaraized_target, reference_data[prediction_column[0]])
                result.reference_metrics.roc_curve = {"fpr": fpr.tolist(), "tpr": tpr.tolist(), "thrs": thrs.tolist()}

                pr, rcl, thrs = metrics.precision_recall_curve(binaraized_target, reference_data[prediction_column[0]])
                result.reference_metrics.pr_curve = {"pr": pr.tolist(), "rcl": rcl.tolist(), "thrs": thrs.tolist()}

                pr_table = []
                step_size = 0.05
                binded = list(zip(binaraized_target["target"].tolist(), reference_data[prediction_column[0]].tolist()))
                binded.sort(key=lambda item: item[1], reverse=True)
                data_size = len(binded)
                target_class_size = sum([x[0] for x in binded])
                offset = max(round(data_size * step_size), 1)

                for step in np.arange(offset, data_size + offset, offset):
                    count = min(step, data_size)
                    prob = round(binded[min(step, data_size - 1)][1], 2)
                    top = round(100.0 * min(step, data_size) / data_size, 1)
                    tp = sum([x[0] for x in binded[: min(step, data_size)]])
                    fp = count - tp
                    precision = round(100.0 * tp / count, 1)
                    recall = round(100.0 * tp / target_class_size, 1)
                    pr_table.append([top, int(count), prob, int(tp), int(fp), precision, recall])

                result.reference_metrics.pr_table = pr_table

            else:
                binaraized_target = pd.DataFrame(binaraized_target)
                binaraized_target.columns = prediction_column

                result.reference_metrics.roc_curve = {}
                result.reference_metrics.pr_curve = {}
                result.reference_metrics.pr_table = {}

                for label in prediction_column:
                    fpr, tpr, thrs = metrics.roc_curve(binaraized_target[label], reference_data[label])
                    result.reference_metrics.roc_curve[label] = {
                        "fpr": fpr.tolist(),
                        "tpr": tpr.tolist(),
                        "thrs": thrs.tolist(),
                    }

                    pr, rcl, thrs = metrics.precision_recall_curve(binaraized_target[label], reference_data[label])
                    result.reference_metrics.pr_curve[label] = {
                        "pr": pr.tolist(),
                        "rcl": rcl.tolist(),
                        "thrs": thrs.tolist(),
                    }

                    pr_table = []
                    step_size = 0.05
                    binded = list(zip(binaraized_target[label].tolist(), reference_data[label].tolist()))
                    binded.sort(key=lambda item: item[1], reverse=True)
                    data_size = len(binded)
                    target_class_size = sum([x[0] for x in binded])
                    offset = max(round(data_size * step_size), 1)

                    for step in np.arange(offset, data_size + offset, offset):
                        count = min(step, data_size)
                        prob = round(binded[min(step, data_size - 1)][1], 2)
                        top = round(100.0 * min(step, data_size) / data_size, 1)
                        tp = sum([x[0] for x in binded[: min(step, data_size)]])
                        fp = count - tp
                        precision = round(100.0 * tp / count, 1)
                        recall = round(100.0 * tp / target_class_size, 1)
                        pr_table.append([top, int(count), prob, int(tp), int(fp), precision, recall])
                    result.reference_metrics.pr_table[label] = pr_table

            if current_data is not None:
                current_data.replace([np.inf, -np.inf], np.nan, inplace=True)
                current_data.dropna(axis=0, how="any", inplace=True, subset=target_and_preds)

                binaraized_target = (current_data[target_column].values.reshape(-1, 1) == prediction_column).astype(int)

                array_prediction = current_data[prediction_column].to_numpy()

                if len(prediction_column) > 2:
                    prediction_ids = np.argmax(array_prediction, axis=-1)
                    prediction_labels = [prediction_column[x] for x in prediction_ids]

                else:
                    maper = {True: prediction_column[0], False: prediction_column[1]}
                    prediction_labels = (current_data[prediction_column[0]] >= classification_threshold).map(maper)

                # calculate quality metrics
                roc_auc = metrics.roc_auc_score(binaraized_target, array_prediction, average="macro")
                log_loss = metrics.log_loss(binaraized_target, array_prediction)
                accuracy_score = metrics.accuracy_score(current_data[target_column], prediction_labels)
                avg_precision = metrics.precision_score(current_data[target_column], prediction_labels, average="macro")
                avg_recall = metrics.recall_score(current_data[target_column], prediction_labels, average="macro")
                avg_f1 = metrics.f1_score(current_data[target_column], prediction_labels, average="macro")

                # calculate class support and metrics matrix
                metrics_matrix = metrics.classification_report(
                    current_data[target_column], prediction_labels, output_dict=True
                )

                roc_aucs = None

                if len(prediction_column) > 2:
                    roc_aucs = metrics.roc_auc_score(binaraized_target, array_prediction, average=None).tolist()

                # calculate confusion matrix
                conf_matrix = metrics.confusion_matrix(current_data[target_column], prediction_labels)
                # get TP, FP, TN, FN metrics for each class
                confusion_by_classes = calculate_confusion_by_classes(conf_matrix, labels)

                result.current_metrics = ProbClassificationPerformanceMetrics(
                    accuracy=accuracy_score,
                    precision=avg_precision,
                    recall=avg_recall,
                    f1=avg_f1,
                    roc_auc=roc_auc,
                    log_loss=log_loss,
                    metrics_matrix=metrics_matrix,
                    confusion_matrix=ConfusionMatrix(labels=labels, values=conf_matrix.tolist()),
                    roc_aucs=roc_aucs,
                    confusion_by_classes=confusion_by_classes,
                )

                # calculate ROC and PR curves, PR table
                if len(prediction_column) <= 2:
                    binaraized_target = pd.DataFrame(binaraized_target[:, 0])
                    binaraized_target.columns = ["target"]

                    fpr, tpr, thrs = metrics.roc_curve(binaraized_target, current_data[prediction_column[0]])
                    result.current_metrics.roc_curve = {"fpr": fpr.tolist(), "tpr": tpr.tolist(), "thrs": thrs.tolist()}

                    pr, rcl, thrs = metrics.precision_recall_curve(
                        binaraized_target, current_data[prediction_column[0]]
                    )
                    result.current_metrics.pr_curve = {"pr": pr.tolist(), "rcl": rcl.tolist(), "thrs": thrs.tolist()}

                    pr_table = []
                    step_size = 0.05
                    binded = list(
                        zip(binaraized_target["target"].tolist(), current_data[prediction_column[0]].tolist())
                    )
                    binded.sort(key=lambda item: item[1], reverse=True)
                    data_size = len(binded)
                    target_class_size = sum([x[0] for x in binded])
                    offset = max(round(data_size * step_size), 1)

                    for step in np.arange(offset, data_size + offset, offset):
                        count = min(step, data_size)
                        prob = round(binded[min(step, data_size - 1)][1], 2)
                        top = round(100.0 * min(step, data_size) / data_size, 1)
                        tp = sum([x[0] for x in binded[: min(step, data_size)]])
                        fp = count - tp
                        precision = round(100.0 * tp / count, 1)
                        recall = round(100.0 * tp / target_class_size, 1)
                        pr_table.append([top, int(count), prob, int(tp), int(fp), precision, recall])

                    result.current_metrics.pr_table = pr_table

                else:
                    binaraized_target = pd.DataFrame(binaraized_target)
                    binaraized_target.columns = prediction_column

                    result.current_metrics.roc_curve = {}
                    result.current_metrics.pr_curve = {}
                    result.current_metrics.pr_table = {}

                    for label in prediction_column:
                        fpr, tpr, thrs = metrics.roc_curve(binaraized_target[label], current_data[label])
                        result.current_metrics.roc_curve[label] = {
                            "fpr": fpr.tolist(),
                            "tpr": tpr.tolist(),
                            "thrs": thrs.tolist(),
                        }

                        pr, rcl, thrs = metrics.precision_recall_curve(binaraized_target[label], current_data[label])
                        result.current_metrics.pr_curve[label] = {
                            "pr": pr.tolist(),
                            "rcl": rcl.tolist(),
                            "thrs": thrs.tolist(),
                        }

                        pr_table = []
                        step_size = 0.05
                        binded = list(zip(binaraized_target[label].tolist(), current_data[label].tolist()))
                        binded.sort(key=lambda item: item[1], reverse=True)
                        data_size = len(binded)
                        target_class_size = sum([x[0] for x in binded])
                        offset = max(round(data_size * step_size), 1)

                        for step in np.arange(offset, data_size + offset, offset):
                            count = min(step, data_size)
                            prob = round(binded[min(step, data_size - 1)][1], 2)
                            top = round(100.0 * min(step, data_size) / data_size, 1)
                            tp = sum([x[0] for x in binded[: min(step, data_size)]])
                            fp = count - tp
                            precision = round(100.0 * tp / count, 1)
                            recall = round(100.0 * tp / target_class_size, 1)
                            pr_table.append([top, int(count), prob, int(tp), int(fp), precision, recall])

                        result.current_metrics.pr_table[label] = pr_table

        return result



from dataclasses import dataclass
from typing import Dict

from evidently.dashboard import Dashboard
from evidently.dashboard.tabs import CatTargetDriftTab
from evidently.dashboard.tabs import ClassificationPerformanceTab
from evidently.dashboard.tabs import DataDriftTab
from evidently.dashboard.tabs import DataQualityTab
from evidently.dashboard.tabs import NumTargetDriftTab
from evidently.dashboard.tabs import ProbClassificationPerformanceTab
from evidently.dashboard.tabs import RegressionPerformanceTab
from evidently.runner.runner import Runner
from evidently.runner.runner import RunnerOptions


@dataclass
class DashboardRunnerOptions(RunnerOptions):
    dashboard_tabs: Dict[str, Dict[str, object]]


tabs_mapping = dict(
    data_drift=DataDriftTab,
    cat_target_drift=CatTargetDriftTab,
    classification_performance=ClassificationPerformanceTab,
    prob_classification_performance=ProbClassificationPerformanceTab,
    num_target_drift=NumTargetDriftTab,
    regression_performance=RegressionPerformanceTab,
    data_quality=DataQualityTab,
)


class DashboardRunner(Runner):
    def __init__(self, options: DashboardRunnerOptions):
        super().__init__(options)
        self.options = options

    def run(self):
        (reference_data, current_data) = self._parse_data()

        tabs = []

        for tab, params in self.options.dashboard_tabs.items():
            tab_class = tabs_mapping.get(tab, None)
            if tab_class is None:
                raise ValueError(f"Unknown tab {tab}")
            try:
                verbose_level = (
                    int(params.get("verbose_level", None)) if params.get("verbose_level", None) is not None else None
                )
            except ValueError as ex:
                raise ValueError(f"Failed to parse verbose level for tab {tab}") from ex
            include_widgets = params.get("include_widgets", None)
            tabs.append(tab_class(verbose_level=verbose_level, include_widgets=include_widgets))

        dashboard = Dashboard(tabs=tabs, options=self.options.options)
        dashboard.calculate(reference_data, current_data, self.options.column_mapping)
        dashboard.save(self.options.output_path + ".html")



from dataclasses import dataclass
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

DEFAULT_CONF_INTERVAL_SIZE = 1
DEFAULT_CLASSIFICATION_THRESHOLD = 0.5


@dataclass
class QualityMetricsOptions:
    conf_interval_n_sigmas: int = DEFAULT_CONF_INTERVAL_SIZE
    classification_threshold: float = DEFAULT_CLASSIFICATION_THRESHOLD
    cut_quantile: Union[None, Tuple[str, float], Dict[str, Tuple[str, float]]] = None

    def as_dict(self):
        return {
            "conf_interval_n_sigmas": self.conf_interval_n_sigmas,
            "classification_threshold": self.classification_threshold,
            "cut_quantile": self.cut_quantile,
        }

    def get_cut_quantile(self, feature_name: str) -> Optional[Tuple[str, float]]:
        if self.cut_quantile is None:
            return None
        if isinstance(self.cut_quantile, tuple):
            return self.cut_quantile
        if isinstance(self.cut_quantile, dict):
            return self.cut_quantile.get(feature_name, None)
        raise ValueError(
            f"""QualityMetricsOptions.cut_quantile
                                is incorrect type {type(self.cut_quantile)}"""
        )



#!/usr/bin/env python
# coding: utf-8
from ._version import __version__
from ._version import version_info
from .nbextension import _jupyter_nbextension_paths
from .pipeline.column_mapping import ColumnMapping
from .pipeline.column_mapping import TaskType



import dataclasses
import uuid
from typing import List
from typing import Optional
from typing import Union

import pandas as pd

from evidently import ColumnMapping
from evidently.base_metric import InputData
from evidently.base_metric import Metric
from evidently.metric_preset.metric_preset import MetricPreset
from evidently.model.dashboard import DashboardInfo
from evidently.model.widget import AdditionalGraphInfo
from evidently.options import ColorOptions
from evidently.renderers.base_renderer import DetailsInfo
from evidently.suite.base_suite import Display
from evidently.suite.base_suite import Suite
from evidently.suite.base_suite import find_metric_renderer
from evidently.utils.data_operations import DatasetColumns
from evidently.utils.data_operations import process_columns
from evidently.utils.data_preprocessing import create_data_definition
from evidently.utils.generators import BaseGenerator


class Report(Display):
    _inner_suite: Suite
    _columns_info: DatasetColumns
    _first_level_metrics: List[Union[Metric]]
    metrics: List[Union[Metric, MetricPreset, BaseGenerator]]

    def __init__(self, metrics: List[Union[Metric, MetricPreset, BaseGenerator]], options: Optional[List] = None):
        super().__init__(options)
        # just save all metrics and metric presets
        self.metrics = metrics
        self._inner_suite = Suite()
        self._first_level_metrics = []

    def run(
        self,
        *,
        reference_data: Optional[pd.DataFrame],
        current_data: pd.DataFrame,
        column_mapping: Optional[ColumnMapping] = None,
    ) -> None:
        if column_mapping is None:
            column_mapping = ColumnMapping()

        if current_data is None:
            raise ValueError("Current dataset should be present")

        self._columns_info = process_columns(current_data, column_mapping)

        self._inner_suite.verify()

        data_definition = create_data_definition(reference_data, current_data, column_mapping)
        data = InputData(reference_data, current_data, None, None, column_mapping, data_definition)

        # get each item from metrics/presets and add to metrics list
        # do it in one loop because we want to save metrics and presets order
        for item in self.metrics:
            # if the item is a metric generator, then we need to generate metrics and add them to the report
            if isinstance(item, BaseGenerator):
                for metric in item.generate(columns_info=self._columns_info):
                    if isinstance(metric, Metric):
                        self._first_level_metrics.append(metric)
                        self._inner_suite.add_metric(metric)

                    else:
                        # if generated item is not a metric, raise an error
                        raise ValueError(f"Incorrect metric type in generator {item}")

            elif isinstance(item, MetricPreset):
                metrics = []

                for metric_item in item.generate_metrics(data=data, columns=self._columns_info):
                    if isinstance(metric_item, BaseGenerator):
                        metrics.extend(metric_item.generate(columns_info=self._columns_info))

                    else:
                        metrics.append(metric_item)

                for metric in metrics:
                    self._first_level_metrics.append(metric)
                    self._inner_suite.add_metric(metric)

            elif isinstance(item, Metric):
                self._first_level_metrics.append(item)
                self._inner_suite.add_metric(item)

            else:
                raise ValueError("Incorrect item instead of a metric or metric preset was passed to Report")
        curr_add, ref_add = self._inner_suite.create_additional_features(current_data, reference_data, data_definition)
        data = InputData(reference_data, current_data, ref_add, curr_add, column_mapping, data_definition)
        self._inner_suite.run_calculate(data)

    def as_dict(self) -> dict:
        metrics = []

        for metric in self._first_level_metrics:
            renderer = find_metric_renderer(type(metric), self._inner_suite.context.renderers)
            metrics.append({"metric": metric.get_id(), "result": renderer.render_json(metric)})

        return {
            "metrics": metrics,
        }

    def _build_dashboard_info(self):
        metrics_results = []
        additional_graphs = []

        color_options = self.options_provider.get(ColorOptions)

        for test in self._first_level_metrics:
            renderer = find_metric_renderer(type(test), self._inner_suite.context.renderers)
            # set the color scheme from the report for each render
            renderer.color_options = color_options
            html_info = renderer.render_html(test)

            for info_item in html_info:
                for additional_graph in info_item.get_additional_graphs():
                    if isinstance(additional_graph, AdditionalGraphInfo):
                        additional_graphs.append(DetailsInfo("", additional_graph.params, additional_graph.id))

                    else:
                        additional_graphs.append(DetailsInfo("", additional_graph, additional_graph.id))

            metrics_results.extend(html_info)

        return (
            "evidently_dashboard_" + str(uuid.uuid4()).replace("-", ""),
            DashboardInfo("Report", widgets=[result for result in metrics_results]),
            {
                f"{item.id}": dataclasses.asdict(item.info) if dataclasses.is_dataclass(item.info) else item.info
                for item in additional_graphs
            },
        )



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from .color_scheme import BERLIN_AUTUMN_COLOR_OPTIONS
from .color_scheme import KARACHI_SUNRISE_COLOR_OPTIONS
from .color_scheme import NIGHTOWL_COLOR_OPTIONS
from .color_scheme import SOLARIZED_COLOR_OPTIONS
from .color_scheme import ColorOptions
from .data_drift import DataDriftOptions
from .quality_metrics import QualityMetricsOptions

TypeParam = TypeVar("TypeParam")


class OptionsProvider:
    _options: Dict[Type, Any]

    def __init__(self):
        self._options = {}

    def add(self, options):
        self._options[type(options)] = options

    def get(self, options_type: Type[TypeParam]) -> TypeParam:
        return self._options.get(options_type, options_type())



import pandas as pd

from evidently.base_metric import ColumnName
from evidently.base_metric import additional_feature
from evidently.features.generated_features import GeneratedFeature
from evidently.utils.data_preprocessing import DataDefinition


class TextLength(GeneratedFeature):
    def __init__(self, column_name: str):
        self.column_name = column_name

    def generate_feature(self, data: pd.DataFrame, data_definition: DataDefinition) -> pd.DataFrame:
        def text_len(s):
            if s is None:
                return 0
            return len(s)

        return pd.DataFrame(dict([(self.column_name, data[self.column_name].apply(text_len))]))

    def feature_name(self) -> ColumnName:
        return additional_feature(self, self.column_name)


def text_length(column_name: str) -> ColumnName:
    return additional_feature(TextLength(column_name), f"{column_name}")



import warnings

import evidently.model_profile.sections
from evidently.model_profile.sections import *

__path__ = evidently.model_profile.sections.__path__  # type: ignore

warnings.warn(
    "'import evidently.profile_sections' is deprecated, use 'import evidently.model_profile.sections'"
)



#!/usr/bin/env python
# coding: utf-8

import json
from typing import Optional

import numpy as np
import pandas as pd
import plotly.graph_objs as go

from evidently import ColumnMapping
from evidently.analyzers.regression_performance_analyzer import (
    RegressionPerformanceAnalyzer,
)
from evidently.dashboard.widgets.widget import Widget
from evidently.model.widget import BaseWidgetInfo
from evidently.options import ColorOptions


class RegPredActualTimeWidget(Widget):
    def __init__(self, title: str, dataset: str = "reference"):
        super().__init__(title)
        self.dataset = dataset  # reference or current

    def analyzers(self):
        return [RegressionPerformanceAnalyzer]

    def calculate(
        self,
        reference_data: pd.DataFrame,
        current_data: Optional[pd.DataFrame],
        column_mapping: ColumnMapping,
        analyzers_results,
    ) -> Optional[BaseWidgetInfo]:
        color_options = self.options_provider.get(ColorOptions)
        results = RegressionPerformanceAnalyzer.get_results(analyzers_results)
        results_utility_columns = results.columns.utility_columns

        if (
            results_utility_columns.target is None
            or results_utility_columns.prediction is None
        ):
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires 'target' and 'prediction' columns"
                )

            return None

        if self.dataset == "current":
            dataset_to_plot = (
                current_data.copy(deep=False) if current_data is not None else None
            )

        else:
            dataset_to_plot = reference_data.copy(deep=False)

        if dataset_to_plot is None:
            if self.dataset == "reference":
                raise ValueError(
                    f"Widget [{self.title}] requires reference dataset but it is None"
                )

            return None

        dataset_to_plot.replace([np.inf, -np.inf], np.nan, inplace=True)
        dataset_to_plot.dropna(
            axis=0,
            how="any",
            inplace=True,
            subset=[results_utility_columns.target, results_utility_columns.prediction],
        )

        # make plots
        pred_actual_time = go.Figure()

        target_trace = go.Scatter(
            x=dataset_to_plot[results_utility_columns.date]
            if results_utility_columns.date
            else dataset_to_plot.index,
            y=dataset_to_plot[results_utility_columns.target],
            mode="lines",
            name="Actual",
            marker=dict(size=6, color=color_options.secondary_color),
        )

        pred_trace = go.Scatter(
            x=dataset_to_plot[results_utility_columns.date]
            if results_utility_columns.date
            else dataset_to_plot.index,
            y=dataset_to_plot[results_utility_columns.prediction],
            mode="lines",
            name="Predicted",
            marker=dict(size=6, color=color_options.primary_color),
        )

        zero_trace = go.Scatter(
            x=dataset_to_plot[results_utility_columns.date]
            if results_utility_columns.date
            else dataset_to_plot.index,
            y=[0] * dataset_to_plot.shape[0],
            mode="lines",
            opacity=0.5,
            marker=dict(
                size=6,
                color=color_options.zero_line_color,
            ),
            showlegend=False,
        )

        pred_actual_time.add_trace(target_trace)
        pred_actual_time.add_trace(pred_trace)
        pred_actual_time.add_trace(zero_trace)

        pred_actual_time.update_layout(
            xaxis_title="Timestamp" if results_utility_columns.date else "Index",
            yaxis_title="Value",
            legend=dict(
                orientation="h", yanchor="bottom", y=1.02, xanchor="right", x=1
            ),
        )

        pred_actual_time_json = json.loads(pred_actual_time.to_json())

        return BaseWidgetInfo(
            title=self.title,
            type="big_graph",
            size=1,
            params={
                "data": pred_actual_time_json["data"],
                "layout": pred_actual_time_json["layout"],
            },
        )
